# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Iterable
from dataclasses import dataclass
from enum import Enum
from typing import Any

import pandas as pd
import polars as pl
import sqlalchemy as sa

import pydiverse.common as pdc
from pydiverse.common import Dtype
from pydiverse.pipedag.backend.table.sql.ddl import (
    CreateTableWithSuffix,
    LockSourceTable,
    LockTable,
)
from pydiverse.pipedag.backend.table.sql.hooks import PandasTableHook, PolarsTableHook
from pydiverse.pipedag.backend.table.sql.reflection import PipedagDB2Reflection
from pydiverse.pipedag.backend.table.sql.sql import SQLTableStore
from pydiverse.pipedag.container import Schema, Table
from pydiverse.pipedag.materialize.details import (
    BaseMaterializationDetails,
    resolve_materialization_details_label,
)

_TABLE_SPACE_KEYWORD_MAP = {
    "table_space_data": "IN",
    "table_space_index": "INDEX IN",
    "table_space_long": "LONG IN",
}


class IBMDB2CompressionTypes(str, Enum):
    """
    Enum for the compression methods for IBM DB2 tables. ``VALUE COMPRESSION`` can
    be combined with one of the other compression methods.

    See the `IBM DB2 documentation\
     <https://www.ibm.com/docs/en/db2/12.1.0?topic=compression-table>`_
    for more details.
    """

    NO_COMPRESSION = ""
    ROW_COMPRESSION = "COMPRESS YES"
    STATIC_ROW_COMPRESSION = "COMPRESS YES STATIC"
    ADAPTIVE_ROW_COMPRESSION = "COMPRESS YES ADAPTIVE"
    VALUE_COMPRESSION = "VALUE COMPRESSION"


@dataclass(frozen=True)
class IBMDB2MaterializationDetails(BaseMaterializationDetails):
    """
    Materialization details specific to IBM DB2.

    :param compression: Specify the compression methods to be applied to the table.
        For possible values see :py:class:`IBMDB2CompressionTypes`.
    :param table_space_data: The DB2 table space where the data is  stored.
    :param table_space_index: The DB2 table space where the partitioned index is stored.
    :param table_space_long: The DB2 table spaces where the values of any long columns
        are stored.

    Example
    -------

    .. code-block:: yaml

      materialization_details:
        __any__:
          compression: ["COMPRESS YES ADAPTIVE", "VALUE COMPRESSION"]
          table_space_data: "USERSPACE1"
        no_compression:
          # user-defined tag. Inherits table_space_data from __any__
          # but overwrites compression.
          compression: ""
    """

    def __post_init__(self):
        object.__setattr__(
            self,
            "compression",
            IBMDB2CompressionTypes(self.compression)
            if isinstance(self.compression, str)
            else [IBMDB2CompressionTypes(c) for c in self.compression]
            if self.compression is not None
            else None,
        )

    compression: IBMDB2CompressionTypes | list[IBMDB2CompressionTypes] | None = None
    table_space_data: str | None = None
    table_space_index: str | None = None
    table_space_long: str | list[str] | None = None


class IBMDB2TableStore(SQLTableStore):
    """
    SQLTableStore that supports `IBM Db2 <https://www.ibm.com/products/db2>`_.
    Requires `ibm-db-sa <https://pypi.org/project/ibm-db-sa/>`_ to be installed.

    Takes the same arguments as
    :py:class:`SQLTableStore <pydiverse.pipedag.backend.table.SQLTableStore>`.

    Supports the :py:class:`IBMDB2MaterializationDetails\
    <pydiverse.pipedag.backend.table.sql.dialects.ibm_db2.IBMDB2MaterializationDetails>`
    materialization details.
    """

    _dialect_name = "ibm_db_sa"

    def _default_isolation_level(self):
        return "UNCOMMITTED READ"

    def lock_table(self, table: Table | str, schema: Schema | str, conn: Any = None) -> list:
        """
        For some dialects, it might be beneficial to lock a table before writing to it.
        """
        stmt = LockTable(table.name if isinstance(table, Table) else table, schema)
        if conn is not None:
            self.execute(stmt, conn=conn)
        return [stmt]

    def lock_source_table(self, table: Table | str, schema: Schema | str, conn: Any = None) -> list:
        """
        For some dialects, it might be beneficial to lock source tables before reading.
        """
        stmt = LockSourceTable(table.name if isinstance(table, Table) else table, schema)
        if conn is not None:
            self.execute(stmt, conn=conn)
        return [stmt]

    def dialect_requests_empty_creation(self, table: Table, is_sql: bool) -> bool:
        if is_sql:
            # IBM DB2 does not support CREATE TABLE AS SELECT without INSERT INTO
            return True
        else:
            label = resolve_materialization_details_label(table)
            return (
                (label is not None and len(label.strip()) > 0)
                or table.nullable is not None
                or table.non_nullable is not None
                or (table.primary_key is not None and len(table.primary_key) > 0)
            )

    def get_forced_nullability_columns(
        self, table: Table, table_cols: Iterable[str], report_nullable_cols=False
    ) -> tuple[list[str], list[str]]:
        # ibm_db2 dialect has literals as non-nullable types by default, so we also need
        # the list of nullable columns to fix
        nullable_cols, non_nullable_cols = self._process_table_nullable_parameters(table, table_cols)
        # add primery key columns to non_nullable_cols
        if table.primary_key:
            primary_key = table.primary_key if isinstance(table.primary_key, list) else [table.primary_key]
            non_nullable_cols += primary_key
            nullable_cols = [col for col in nullable_cols if col not in table.primary_key]
        return nullable_cols, non_nullable_cols

    def add_indexes_and_set_nullable(
        self,
        table: Table,
        schema: Schema,
        *,
        on_empty_table: bool | None = None,
        table_cols: Iterable[str] | None = None,
    ):
        super().add_indexes_and_set_nullable(table, schema, on_empty_table=on_empty_table, table_cols=table_cols)
        table_name = self.engine.dialect.identifier_preparer.quote(table.name)
        schema_name = self.engine.dialect.identifier_preparer.quote_schema(schema.get())
        query = (
            f"CALL SYSPROC.ADMIN_CMD('RUNSTATS ON TABLE {schema_name}.{table_name}"
            f" ON ALL COLUMNS WITH DISTRIBUTION ON ALL COLUMNS AND UNSAMPLED"
            f" DETAILED INDEXES ALL SET PROFILE');"
        )
        self.execute(query)

    def resolve_alias(self, table: Table, stage_name: str) -> tuple[str, str]:
        # The base implementation already takes care of converting Table objects
        # based on ExternalTableReference objects to string table name and schema.
        # For normal Table objects, it needs the stage schema name.
        table_name, schema = super().resolve_alias(table, stage_name)
        return PipedagDB2Reflection.resolve_alias(self.engine, table_name, schema)

    def check_materialization_details_supported(self, label: str | None) -> None:
        _ = label
        return

    def _set_materialization_details(self, materialization_details: dict[str, dict[str | list[str]]] | None) -> None:
        self.materialization_details = IBMDB2MaterializationDetails.create_materialization_details_dict(
            materialization_details,
            self.strict_materialization_details,
            self.default_materialization_details,
            self.logger,
        )

    def _get_compression(self, materialization_details_label: str | None) -> str | list[str] | None:
        compression: IBMDB2CompressionTypes | list[IBMDB2CompressionTypes] | None = (
            IBMDB2MaterializationDetails.get_attribute_from_dict(
                self.materialization_details,
                materialization_details_label,
                self.default_materialization_details,
                "compression",
                self.strict_materialization_details,
                self.logger,
            )
        )
        if isinstance(compression, list):
            return [c.value for c in compression]
        if compression is not None:
            return compression.value

    def _get_table_spaces(self, materialization_details_label: str | None) -> dict[str, str]:
        return {
            f"table_space_{st}": IBMDB2MaterializationDetails.get_attribute_from_dict(
                self.materialization_details,
                materialization_details_label,
                self.default_materialization_details,
                f"table_space_{st}",
                self.strict_materialization_details,
                self.logger,
            )
            for st in ("data", "index", "long")
        }

    def get_create_table_suffix(self, materialization_details_label: str | None) -> str:
        table_spaces = self._get_table_spaces(materialization_details_label)
        table_space_suffix = " ".join(
            f"{_TABLE_SPACE_KEYWORD_MAP[stype]} {self.engine.dialect.identifier_preparer.quote(sname)}"
            for stype, sname in table_spaces.items()
            if sname
        )
        compression = self._get_compression(materialization_details_label)
        if isinstance(compression, str):
            compression = [compression]
        elif compression is None:
            compression = []
        compression_suffix = " ".join(compression)
        return " ".join((table_space_suffix, compression_suffix))

    def _get_all_objects_in_schema(self, schema: Schema) -> dict[str, Any]:
        return PipedagDB2Reflection.get_all_objects(self.engine, schema.get())


class DataframeIbmDb2TableHook:
    @classmethod
    def dialect_has_adbc_driver(cls):
        # No ADBC driver for IBM DB2 is available.
        return False

    @classmethod
    def dialect_supports_connectorx(cls):
        # ConnectorX (used by Polars read_database_uri) does not support IBM DB2
        return False

    @classmethod
    def _get_dialect_dtypes(
        cls, dtypes: dict[str, Dtype], table: Table[pd.DataFrame | pl.DataFrame]
    ) -> dict[str, sa.types.TypeEngine]:
        # Default string target is CLOB which can't be used for indexing.
        # -> Convert indexed string columns to VARCHAR(256)
        index_columns = set()
        if indexes := table.indexes:
            index_columns |= {col for index in indexes for col in index}
        if primary_key := table.primary_key:
            index_columns |= set(primary_key)

        sql_dtypes = (
            {name: dtype.to_sql() for name, dtype in dtypes.items()}
            | {
                name: (sa.String(length=256) if name in index_columns else sa.String(length=32_672))
                for name, dtype in dtypes.items()
                if dtype == pdc.String()
            }
            | {
                name: sa.String(length=max(len(c.encode("utf")) for c in dtype.categories))
                for name, dtype in dtypes.items()
                if isinstance(dtype, pdc.Enum)
            }
        )

        if table.type_map:
            sql_dtypes.update(table.type_map)
        return sql_dtypes

    @classmethod
    def _dialect_create_empty_table(
        cls,
        store: SQLTableStore,
        table: Table[pd.DataFrame | pl.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
    ):
        suffix = store.get_create_table_suffix(resolve_materialization_details_label(table))
        store.execute(
            CreateTableWithSuffix(
                table.name,
                schema,
                dtypes,
                table.nullable,
                table.non_nullable,
                suffix,
            )
        )


@IBMDB2TableStore.register_table(pd)
class PandasTableHook(DataframeIbmDb2TableHook, PandasTableHook):
    pass


@IBMDB2TableStore.register_table(pd)
class PolarsTableHook(DataframeIbmDb2TableHook, PolarsTableHook):
    @classmethod
    def dialect_supports_polars_native_read(cls):
        # Polars read_database is utterly broken for IBM DB2.
        # Polars read_database_uri uses connectorx which does not support IBM DB2.
        return False



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import pytest
import sqlalchemy as sa

from pydiverse.pipedag import (
    ConfigContext,
    Flow,
    GroupNode,
    PipedagConfig,
    Stage,
    Table,
    input_stage_versions,
    materialize,
)
from pydiverse.pipedag.context import StageLockContext

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import (
    ALL_INSTANCES,
    ORCHESTRATION_INSTANCES,
    with_instances,
)
from tests.util import swallowing_raises
from tests.util import tasks_library as m

try:
    from sqlalchemy import Alias
except ImportError:
    # For compatibility with sqlalchemy < 2.0
    from sqlalchemy.sql import alias as Alias

pytestmark = [with_instances(ALL_INSTANCES, ORCHESTRATION_INSTANCES)]


def test_input_versions_literals():
    """
    Check that all json serializable literals get stored and retrieved
    correctly from the table / metadata store.
    """

    @input_stage_versions(pass_args=["value"])
    def noop(transaction_table_dict, other_table_dict, value):
        _ = transaction_table_dict, other_table_dict
        return value

    def materialize_and_retrieve(value):
        v = noop(value=value)
        return m.assert_equal(value, v)

    with Flow("flow") as f:
        with Stage("stage"):
            materialize_and_retrieve(None)

            materialize_and_retrieve(0)
            materialize_and_retrieve(1)

            # max 32bit
            materialize_and_retrieve(-2147483648)
            materialize_and_retrieve(2147483647)

            # max 64bit
            materialize_and_retrieve(-9223372036854775808)
            materialize_and_retrieve(9223372036854775807)

            # in python json we even have infinite precision integers:
            materialize_and_retrieve(-9223372036854775809)
            materialize_and_retrieve(9223372036854775808)

            materialize_and_retrieve(3.14)
            materialize_and_retrieve(-2.71)

            materialize_and_retrieve("")
            materialize_and_retrieve("a string")
            materialize_and_retrieve("a" * 256)

            materialize_and_retrieve([0, 1, 2, 3])

            materialize_and_retrieve({"0": 0, "1": 1})

            materialize_and_retrieve(
                {
                    "key": [0, 1, 2, 3],
                    "numbers": {
                        "zero": 0,
                        "one": 1,
                    },
                    "nested_lists": [[[None]]],
                }
            )

    assert f.run().successful


def test_input_versions_table():
    run = 1
    val = 12

    @input_stage_versions(input_type=pd.DataFrame, ordering_barrier=False)
    def validate_stage(tbls: dict[str, pd.DataFrame], other_tbls: dict[str, pd.DataFrame]):
        if run > 1:
            assert len(tbls) == 1
            assert len(other_tbls) == 1
            x = tbls["x"]
            y = other_tbls["x"]
            pd.testing.assert_frame_equal(x, y)

    @materialize
    def pd_dataframe(data: dict[str, list]):
        return Table(pd.DataFrame(data), name="x")

    def get_flow():
        with Flow("flow") as f:
            with Stage("stage"):
                x = pd_dataframe({"a": [val], "b": [24]})
                validate_stage(x)
                m.assert_table_equal(x, x)
                with GroupNode(ordering_barrier=True):
                    validate_stage()
        return f

    with StageLockContext():
        f = get_flow()
        assert f.run().successful
        run += 1
        f = get_flow()
        assert f.run().successful
        run += 1
        val = -5
        f = get_flow()
        with swallowing_raises(
            AssertionError,
            match=r"\[left\]:[ \t]*\[-5(\.0)?\]\n[ \t]*\[right\]:[ \t]*\[12(\.0)?\]",
        ):
            f.run()


@with_instances("postgres")
def test_input_versions_blob():
    run = 1
    val = 12

    @input_stage_versions(input_type=pd.DataFrame)
    def validate_stage(
        tbls: dict[str, Alias],
        other_tbls: dict[str, Alias],
        blobs: dict[str, dict],
        other_blobs: dict[str, dict],
    ):
        if run > 1:
            assert len(tbls) == 0
            assert len(other_tbls) == 0
            assert len(blobs) == 3
            assert len(other_blobs) == 3
            # this exception fails on different input because the hashes change and thus
            # the file name changes
            assert not any(isinstance(x, Exception) for x in other_blobs.values()), "expected to fail on third call"
            assert blobs == other_blobs

    def get_flow():
        with Flow("flow") as f:
            with Stage("stage_0"):
                x = m.object_blob({"a": val, "b": 24})

            with Stage("stage_1"):
                y = m.as_blob(x)
                m.assert_blob_equal(x, y)
                one = m.one()
                one_blob = m.as_blob(m.one())
                m.assert_equal(one, one_blob)
                validate_stage([x, y, one_blob])
        return f

    with StageLockContext():
        f = get_flow()
        assert f.run().successful
        run += 1
        f = get_flow()
        assert f.run().successful
        run += 1
        val = -5
        f = get_flow()
        with swallowing_raises(AssertionError, match=r"expected to fail on third call"):
            f.run()


@with_instances("mssql")  # mssql can do cross-database queries
@pytest.mark.parametrize("per_user", [True, False])
def test_input_versions_other_instance_table(per_user):
    run = 1
    val = 12

    @input_stage_versions(input_type=sa.Table, lazy=True, lock_source_stages=False)
    def join_across_stage_versions(
        tbls: dict[str, Alias],
        other_tbls: dict[str, Alias],
        other_cfg: ConfigContext,
    ):
        if run > 1:
            # make a cross-database query to combine tables of both instances
            other_database = other_cfg.store.table_store.engine.url.database
            assert ConfigContext.get().store.table_store.engine.url.database != other_database
            other_tbls["x"].original.schema = f"{other_database}.{other_tbls['x'].original.schema}"
            return Table(
                sa.select(tbls["x"].outerjoin(other_tbls["x"], tbls["x"].c.a == other_tbls["x"].c.a)),
                name="res",
            )
        else:
            return Table(tbls["x"], name="res")

    @input_stage_versions(input_type=pd.DataFrame, ordering_barrier=False, lock_source_stages=False)
    def validate_stage(
        tbls: dict[str, pd.DataFrame],
        other_tbls: dict[str, pd.DataFrame],
        other_cfg: ConfigContext,
    ):
        _ = other_cfg
        if run > 1:
            assert 1 <= len({k for k in tbls.keys() if not k.endswith("__copy")}) <= 2
            assert 1 <= len({k for k in other_tbls.keys() if not k.endswith("__copy")}) <= 3
            x = tbls["x"]
            y = other_tbls["x"]
            pd.testing.assert_frame_equal(x, y)

    @materialize(version="1.1")
    def pd_dataframe(data: dict[str, list]):
        return Table(pd.DataFrame(data), name="x")

    def get_flow(other_cfg: ConfigContext):
        with Flow("flow") as f:
            with Stage("stage"):
                x = pd_dataframe({"a": [val], "b": [24]})
                res = join_across_stage_versions(other_cfg)
                validate_stage(x, other_cfg)
                m.assert_table_equal(x, x)
                with GroupNode(ordering_barrier=True):
                    validate_stage(config=other_cfg)
                _ = m.noop(res)
        return f

    cfg = ConfigContext.get()
    cfg2 = PipedagConfig.default.get("mssql_pytsql", per_user=per_user)

    with StageLockContext():
        f = get_flow(cfg2)
        assert cfg.store.table_store.engine.url != cfg2.store.table_store.engine.url
        assert f.run(config=cfg).successful
        assert (
            pd.read_sql_table(
                "x",
                schema=cfg.store.table_store.get_schema(f["stage"].name).get(),
                con=cfg.store.table_store.engine,
            ).iloc[0, 0]
            == val
        )
        run += 1
        f = get_flow(cfg)
        assert f.run(config=cfg2).successful
        assert (
            pd.read_sql_table(
                "x",
                schema=cfg2.store.table_store.get_schema(f["stage"].name).get(),
                con=cfg2.store.table_store.engine,
            ).iloc[0, 0]
            == val
        )
        run += 1
        val = -5
        f = get_flow(cfg)
        with swallowing_raises(AssertionError, match=r"\[left\]:  \[-5\]\n\[right\]: \[12\]"):
            f.run(config=cfg2)
        assert (
            pd.read_sql_table(
                "x",
                schema=cfg2.store.table_store.get_schema(f["stage"].transaction_name).get(),
                con=cfg2.store.table_store.engine,
            ).iloc[0, 0]
            == val
        )


@with_instances("postgres")
@pytest.mark.parametrize("per_user", [True, False])
def test_input_versions_other_instance_locking(per_user):
    run = 1
    val = 12

    @input_stage_versions(input_type=pd.DataFrame, lazy=True)
    def join_across_stage_versions(
        tbls: dict[str, pd.DataFrame],
        other_tbls: dict[str, pd.DataFrame],
        other_cfg: ConfigContext,
    ):
        _ = other_cfg
        # other_tbls["x"] might be destroyed by a race condition since we can't keep
        # locks between both runs because we lock the other instance again
        if run > 1 and isinstance(other_tbls["x"], pd.DataFrame):
            return Table(tbls["x"].merge(other_tbls["x"], on="a"), name="res")
        else:
            return Table(tbls["x"], name="res")

    @input_stage_versions(input_type=pd.DataFrame, ordering_barrier=False)
    def validate_stage(
        tbls: dict[str, pd.DataFrame],
        other_tbls: dict[str, pd.DataFrame],
        other_cfg: ConfigContext,
    ):
        _ = other_cfg
        # we cannot make assumptions since we did not lock other instance between the
        # two runs
        _ = other_tbls
        if run > 1:
            assert 1 <= len({k for k in tbls.keys() if not k.endswith("__copy")}) <= 2
            x = tbls["x"]
            assert x.iloc[0, 0] == val

    @materialize
    def pd_dataframe(data: dict[str, list]):
        return Table(pd.DataFrame(data), name="x")

    def get_flow(other_cfg: ConfigContext):
        with Flow("flow") as f:
            with Stage("stage"):
                x = pd_dataframe({"a": [val], "b": [24]})
                res = join_across_stage_versions(other_cfg)
                validate_stage(x, other_cfg)
                m.assert_table_equal(x, x)
                with GroupNode(ordering_barrier=True):
                    validate_stage(config=other_cfg)
                _ = m.noop(res)
        return f

    cfg = ConfigContext.get()
    cfg2 = PipedagConfig.default.get("postgres_unlogged", per_user=per_user)

    with StageLockContext():
        f = get_flow(cfg2)
        assert cfg.store.table_store.engine.url != cfg2.store.table_store.engine.url
        assert f.run(config=cfg).successful
        assert (
            pd.read_sql_table(
                "x",
                schema=cfg.store.table_store.get_schema(f["stage"].name).get(),
                con=cfg.store.table_store.engine,
            ).iloc[0, 0]
            == val
        )

    run += 1

    with StageLockContext():
        f = get_flow(cfg)
        assert f.run(config=cfg2).successful
        assert (
            pd.read_sql_table(
                "x",
                schema=cfg2.store.table_store.get_schema(f["stage"].name).get(),
                con=cfg2.store.table_store.engine,
            ).iloc[0, 0]
            == val
        )

    run += 1
    val = -5

    with StageLockContext():
        f = get_flow(cfg)
        assert f.run(config=cfg2).successful
        assert (
            pd.read_sql_table(
                "x",
                schema=cfg2.store.table_store.get_schema(f["stage"].name).get(),
                con=cfg2.store.table_store.engine,
            ).iloc[0, 0]
            == val
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import TYPE_CHECKING

import structlog

from pydiverse.pipedag._typing import Materializable
from pydiverse.pipedag.context import RunContext, RunContextServer

if TYPE_CHECKING:
    from pydiverse.pipedag import Result, Table, Task
    from pydiverse.pipedag.materialize.cache import TaskCacheInfo
    from pydiverse.pipedag.materialize.metadata import TaskMetadata


class TraceHook:
    def run_init_context_server(self, run_context_server: RunContextServer):
        """
        Called after run context server was initialized.

        The run context server launches a central service in case of multi-node
        execution. Hooks might integrate with it to collect information at a central
        point.
        """
        pass

    def run_init_context(self, run_context: RunContext):
        """
        Called after initialization of run context.

        This happens when entering RunContextServer context.
        """
        pass

    def run_complete(self, result: "Result"):
        """
        Called after completion of pipedag run when result is available.
        """
        pass

    def stage_post_init(self, stage_id: int, success: bool):
        """
        Called after initialization of stage.

        The stage schema can be used for writing tables from this moment on.
        """
        pass

    def stage_pre_commit(self, stage_id: int):
        """
        Called before committing stage.
        """
        pass

    def stage_post_commit(self, stage_id: int, success: bool):
        """
        Called after committing stage.

        Stage commit may include copying remaining tables from cache to transaction
        schema.
        """
        pass

    def task_begin(self, task: "Task"):
        """
        Called at beginning of task execution.

        This typically is just very shortly before task_cache_status. It also can be
        used as the starting point for dematerialization.
        """
        pass

    def task_cache_status(
        self,
        task: "Task",
        input_hash: str,
        cache_fn_hash: str,
        cache_metadata: "TaskMetadata | None" = None,
        cached_output: Materializable | None = None,
        cache_valid: bool | None = None,
        lazy: bool | None = None,
    ):
        """
        Called after it is known whether task is cache valid for non-lazy tasks.

        It is not called in case cache validity check is disabled. For lazy tasks,
        it is also called for technical reasons because the cache_fn_hash may be
        retrieved from the cache in order to avoid recomputation in some special
        cases.
        """
        pass

    def task_pre_call(self, task: "Task"):
        """
        Called before call of task function.

        This is the same moment when dematerialization completed.
        """
        pass

    def task_post_call(self, task: "Task"):
        """
        Called after call of task function.

        This is the same moment when materialization begins.
        """
        pass

    def task_complete(self, task: "Task", result: "Result"):
        """
        Called after task completion.

        This is the same moment when materialization completes.
        All query related callbacks of this task are called before this.
        """
        pass

    def query_cache_status(
        self,
        task: "Task",
        table: "Table",
        task_cache_info: "TaskCacheInfo",
        query_hash: str,
        query_str: str,
        cache_metadata: "TaskMetadata | None" = None,
        cache_valid: bool | None = None,
    ):
        """
        Called after it is known whether query is cache valid.

        Task caching information is available in task_cache_info.
        In addition, query_str and table.assumed_dependencies are used for computing
        query_hash.
        """
        pass

    def query_complete(self, task: "Task", table: "Table"):
        """
        Called after query completion in case query is not cache valid.
        """
        pass

    def cache_init_transfer(self, task: "Task", table: "Table"):
        """
        Called when preparing transfer from cached table to transactions schema.
        """
        pass

    def cache_pre_transfer(self, table: "Table"):
        """
        Called before transferring cached table to transaction schema.

        This function may be called in context of another thread or process.
        """
        pass

    def cache_post_transfer(self, table: "Table"):
        """
        Called after transferring cached table to transaction schema.

        This function may be called in context of another thread or process.
        """
        pass


class PrintTraceHook(TraceHook):
    def __init__(self):
        self.logger = structlog.get_logger(logger_name=type(self).__name__)
        self.run_context = None
        self.run_context_server = None

    def run_init_context_server(self, run_context_server: RunContextServer):
        self.run_context_server = run_context_server
        self.logger.debug("run_init_context_server", run_context_server=run_context_server)

    def run_init_context(self, run_context: RunContext):
        self.run_context = run_context
        self.logger.debug("run_init_context", run_context=run_context)

    def run_complete(self, result: "Result"):
        self.logger.debug("run_complete", result=result)

    def stage_post_init(self, stage_id: int, success: bool):
        self.logger.debug(
            "stage_post_init",
            stage_id=stage_id,
            stage=list(self.run_context.flow.stages.values())[stage_id],
            success=success,
        )

    def stage_pre_commit(self, stage_id: int):
        self.logger.debug(
            "stage_pre_commit",
            stage_id=stage_id,
            stage=list(self.run_context.flow.stages.values())[stage_id],
        )

    def stage_post_commit(self, stage_id: int, success: bool):
        self.logger.debug("stage_post_commit", stage_id=stage_id, success=success)

    def task_begin(self, task: "Task"):
        self.logger.debug("task_begin", task=task)

    def task_cache_status(
        self,
        task: "Task",
        input_hash: str,
        cache_fn_hash: str,
        cache_metadata: "TaskMetadata | None" = None,
        cached_output: Materializable | None = None,
        cache_valid: bool | None = None,
        lazy: bool | None = None,
    ):
        self.logger.debug(
            "task_cache_status",
            task=task,
            input_hash=input_hash,
            cache_fn_hash=cache_fn_hash,
            cache_metadata=cache_metadata,
            cached_output=cached_output,
            cache_valid=cache_valid,
            lazy=lazy,
        )

    def task_pre_call(self, task: "Task"):
        self.logger.debug("task_pre_call", task=task)

    def task_post_call(self, task: "Task"):
        self.logger.debug("task_post_call", task=task)

    def task_complete(self, task: "Task", result: "Result"):
        self.logger.debug("task_complete", task=task, result=result)

    def query_cache_status(
        self,
        task: "Task",
        table: "Table",
        task_cache_info: "TaskCacheInfo",
        query_hash: str,
        query_str: str,
        cache_metadata: "TaskMetadata | None" = None,
        cache_valid: bool | None = None,
    ):
        self.logger.debug(
            "query_cache_status",
            task=task,
            table=table,
            detail=table.assumed_dependencies,
            task_cache_info=task_cache_info,
            query_hash=query_hash,
            query=query_str,
            cache_metadata=cache_metadata,
            cache_valid=cache_valid,
        )

    def query_complete(self, task: "Task", table: "Table"):
        self.logger.debug("query_complete", task=task, table=table)

    def cache_init_transfer(self, task: "Task", table: "Table"):
        self.logger.debug("cache_init_transfer", task=task, table=table)

    def cache_pre_transfer(self, table: "Table"):
        self.logger.debug("cache_pre_transfer", table=table)

    def cache_post_transfer(self, table: "Table"):
        self.logger.debug("cache_post_transfer", table=table)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import itertools
import logging
import os
from collections import defaultdict
from pathlib import Path

import pytest

from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag.util.testing_s3 import initialize_test_s3_bucket

# Load the `run_with_instances` fixture, so it gets applied to all tests
from tests.fixtures.instances import INSTANCE_MARKS, fixture_run_with_instance
from tests.util.dask_patch import *  # noqa: F401,F403

_ = fixture_run_with_instance


pytest_plugins = ["tests.parallelize.plugin"]


# Setup


def setup_environ():
    if "PIPEDAG_CONFIG" not in os.environ:
        os.environ["PIPEDAG_CONFIG"] = str(Path(__file__).parent / "resources")
    os.environ["POSTGRES_USERNAME"] = "sa"
    os.environ["POSTGRES_PASSWORD"] = "Pydiverse23"
    os.environ["MSSQL_USERNAME"] = "sa"
    os.environ["MSSQL_PASSWORD"] = "PydiQuant27"

    os.environ["PYDIVERSE_PIPEDAG_PYTEST"] = "1"
    os.environ["AWS_ACCESS_KEY_ID"] = "minioadmin"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "minioadmin"
    os.environ["AWS_ENDPOINT_URL"] = "http://localhost:9000"

    from fsspec.config import conf  # fsspec’s global config

    conf.setdefault("s3", {})
    conf["s3"].update(
        {
            "key": "minioadmin",
            "secret": "minioadmin",
            "client_kwargs": {
                "endpoint_url": "http://localhost:9000",
            },
            # ─── botocore.config.Config parameters go here ────
            "config_kwargs": {
                "connect_timeout": 3,
                "read_timeout": 5,
                "retries": {"max_attempts": 2, "mode": "standard"},
                "s3": {"addressing_style": "path"},  # MinIO needs path-style
            },
        }
    )


setup_environ()

log_level = (
    logging.ERROR
    if os.environ.get("ERROR_ONLY", "0") != "0"
    else logging.INFO
    if os.environ.get("DEBUG", "0") == "0"
    else logging.DEBUG
)
setup_logging(log_level=log_level)


# Pytest Configuration


@pytest.fixture(autouse=True, scope="function")
def structlog_test_info(request):
    """Add testcase information to structlog context"""
    if os.environ.get("DEBUG", "0") == "0" and os.environ.get("LOG_TEST_NAME", "0") == "0":
        yield
        return

    import structlog

    with structlog.contextvars.bound_contextvars(testcase=request.node.name):
        yield


def pytest_generate_tests(metafunc: pytest.Metafunc):
    # Parametrize tests based on `instances` and `skip_instances` mark.
    # Depends on the `fixture_run_with_instance` fixture to be imported at a
    #     global level.

    instances = {}
    kwargs = {}
    if mark := metafunc.definition.get_closest_marker("instances"):
        instances = dict.fromkeys(mark.args)
        kwargs = dict(mark.kwargs)
    if mark := metafunc.definition.get_closest_marker("skip_instances"):
        for instance in mark.args:
            if instance in instances:
                del instances[instance]

    if len(instances):
        params = []
        for instance in instances:
            marks = INSTANCE_MARKS.get(instance, ())
            params.append(
                pytest.param(
                    (instance, kwargs),
                    marks=marks,
                    id=instance,
                )
            )

        metafunc.parametrize("run_with_instance", params, indirect=True)


supported_options = [
    "postgres",  # see default_options
    "snowflake",
    "mssql",
    "ibm_db2",
    "duckdb",
    "pdtransform",
    "ibis",
    "polars",
    "dask",
    "prefect",
    "s3",
]

default_options = ["postgres", "duckdb", "polars"]

sub_backends = dict(duckdb=["parquet_backend"], s3=["parquet_s3_backend"])


def pytest_addoption(parser):
    for opt in supported_options:
        parser.addoption(
            "--" + opt,
            action="store_true",
            default=False,
            help=f"run test that require {opt}",
        )
        if opt in default_options:
            # it seems the pytest parser is not automatically offering --no-flag
            parser.addoption(
                "--no-" + opt,
                action="store_true",
                default=False,
                help=f"run test that require {opt}",
            )


def pytest_collection_modifyitems(config: pytest.Config, items):
    for opt in supported_options:
        if not (config.getoption("--" + opt) or (opt in default_options and not config.getoption("--no-" + opt))):
            all_opts = [opt]
            if opt in sub_backends:
                all_opts.extend(sub_backends[opt])
            for opt2 in all_opts:
                skip = pytest.mark.skip(reason=f"{opt2} not selected")
                for item in items:
                    if (
                        opt2 in item.keywords
                        or any(itm.startswith(opt2 + "-") for itm in item.keywords)
                        or any(itm.endswith("-" + opt2) for itm in item.keywords)
                        or any("-" + opt2 + "-" in itm for itm in item.keywords)
                    ):
                        item.add_marker(skip)


@pytest.hookimpl
def pytest_parallelize_group_items(config, items):
    groups = defaultdict(list)
    auto_group_iter = itertools.cycle([f"auto_{i}" for i in range(os.cpu_count() or 4)])
    for item in items:
        group = "DEFAULT"

        if hasattr(item, "get_closest_marker"):
            if marker := item.get_closest_marker("parallelize"):
                if marker.args:
                    group = marker.args[0]
                else:
                    group = next(auto_group_iter)

        if hasattr(item, "callspec"):
            if instance := item.callspec.params.get("run_with_instance"):
                # instance is a tuple of (instance_name, config_kwargs)
                group = instance[0]

        groups[group].append(item)
    return groups


def pytest_sessionstart(session):
    # initialize the S3 test bucket in case --s3 option was given
    if session.config.getoption("--s3"):
        initialize_test_s3_bucket()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

from pydiverse.pipedag import Flow, Stage
from pydiverse.pipedag.context import ConfigContext, FinalTaskState
from tests.fixtures.instances import with_instances
from tests.test_raw_sql.util import sql_script


# TODO: Extend tests for other backends
@with_instances("mssql")
def test_raw_sql_schema_swap():
    # This test creates various different objects in one schema and then
    # checks if, after swapping the schema, if they are still working correctly.

    instance_name = ConfigContext.get().instance_name
    dir_ = Path(__file__).parent / "scripts" / instance_name / "schema_swap"

    with Flow() as f:
        with Stage("raw_0") as raw_0:
            sql_1 = sql_script("create_objects.sql", dir_)
        with Stage("raw_1"):
            sql_2 = sql_script("check_objects.sql", dir_, input_stage=raw_0, depend=[sql_1])

    f.run()

    # Check that running the flow again results in the cache being used
    for _ in range(2):
        result = f.run()
        assert result.task_states[sql_1] == FinalTaskState.CACHE_VALID
        assert result.task_states[sql_2] == FinalTaskState.CACHE_VALID



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import itertools
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING

from pydiverse.common.util.hashing import stable_hash
from pydiverse.pipedag import Table

if TYPE_CHECKING:
    from pydiverse.pipedag.materialize.materializing_task import MaterializingTask


class ImperativeMaterializationState:
    def __init__(self):
        # every imperatively materialized table is an assumed dependency of
        # subsequently materialized tables of the same task
        self.assumed_dependencies: set[Table] = set()
        # Table(...).materialize() returns dematerialized objects. We need to find the
        # corresponding Table objects for handing returned objects over to consumer
        # tasks.
        self.object_lookup: dict[int, Table] = {}
        self.table_ids: set[int] = set()
        self.auto_suffix_counter = itertools.count()

    def add_table_lookup(self, obj, table: Table):
        self.assumed_dependencies.add(table)
        self.object_lookup[id(obj)] = table
        self.table_ids.add(id(table))


@dataclass(frozen=True)
class TaskCacheInfo:
    task: "MaterializingTask"
    input_hash: str
    cache_fn_hash: str
    cache_key: str
    assert_no_materialization: bool
    force_task_execution: bool

    @cached_property
    def imperative_materialization_state(self):
        """State used by Table.materialize()"""
        return ImperativeMaterializationState()


def task_cache_key(task: "MaterializingTask", input_hash: str, cache_fn_hash: str):
    """Cache key used to judge cache validity of the current task output.

    Also referred to as `task_hash`.

    For lazy objects, this hash isn't used to judge cache validity, instead it
    serves as an identifier to reference a specific task run. This can be the case
    if a task is determined to be cache-valid and the lazy query string is also
    the same, but the task_hash is different from a previous run. Then we can
    compute this combined_cache_key from the task's cache metadata to determine
    which lazy object to use as cache.

    :param task: task for which the cache key is computed
    :param input_hash: hash used for checking whether task is cache invalid due
        to changing input.
    :param cache_fn_hash: same as input_hash but for external inputs which need
        manual cache invalidation function.
    :return: The hash / cache key (str).
    """

    return stable_hash(
        "TASK",
        task.name,
        task.version,
        input_hash,
        cache_fn_hash,
    )


def lazy_table_cache_key(task_hash: str, query_hash: str):
    return stable_hash(
        "LAZY_TABLE",
        task_hash,
        query_hash,
    )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import sqlalchemy as sa

from pydiverse.pipedag import Flow, Stage, materialize


# Define the different tasks our flow consists of
@materialize(version="1.0", nout=2)
def input_tables():
    names = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "name": ["Alice", "Bob", "Charlie"],
        }
    )

    ages = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "age": [20, 40, 60],
        }
    )

    return names, ages


@materialize(lazy=True, input_type=sa.Table)
def join_tables(names, ages):
    return sa.select(names.c.id, names.c.name, ages.c.age).join_from(names, ages, names.c.id == ages.c.id)


@materialize(input_type=pd.DataFrame)
def print_dataframe(df):
    print(df)


def main():
    # Define how the different tasks should be wired
    with Flow("flow") as flow:
        with Stage("inputs"):
            names, ages = input_tables()

        with Stage("features"):
            joined_table = join_tables(names, ages)
            print_dataframe(joined_table)

    # # In case you provide a pipedag.yaml, you can run the flow as simple as:
    # flow.run()

    # run flow with a duckdb configuration in a random temporary directory (this is
    # easier to get started)
    import tempfile

    from pydiverse.pipedag.core.config import create_basic_pipedag_config

    with tempfile.TemporaryDirectory() as temp_dir:
        cfg = create_basic_pipedag_config(
            f"duckdb:///{temp_dir}/db.duckdb",
            disable_stage_locking=True,  # This is special for duckdb
        ).get("default")
        # Execute the flow
        flow.run(config=cfg)


if __name__ == "__main__":
    from pydiverse.common.util.structlog import setup_logging

    setup_logging()  # you can setup the logging and/or structlog libraries as you wish
    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.pipedag import ExternalTableReference, Table, Task
from pydiverse.pipedag.context import ConfigContext, RunContext
from pydiverse.pipedag.core import Subflow
from pydiverse.pipedag.core.result import Result
from pydiverse.pipedag.engine.base import (
    OrchestrationEngine,
)


class SequentialEngine(OrchestrationEngine):
    """Most basic orchestration engine that just executes all tasks sequentially."""

    def run(
        self,
        flow: Subflow,
        ignore_position_hashes: bool = False,
        inputs: dict[Task, ExternalTableReference] | None = None,
        **run_kwargs,
    ):
        run_context = RunContext.get()
        config_context = ConfigContext.get()

        failed_tasks = set()  # type: set[Task]
        results = {}
        exception = None
        inputs = inputs if inputs is not None else {}

        try:
            for task in flow.get_tasks():
                try:
                    if not (set(task.input_tasks) & failed_tasks):
                        task_inputs = {
                            **{
                                in_id: results[in_t]
                                for in_id, in_t in task.input_tasks.items()
                                if in_t in results and in_t not in inputs
                            },
                            **{
                                in_id: Table(inputs[in_t]) for in_id, in_t in task.input_tasks.items() if in_t in inputs
                            },
                        }

                        results[task] = task.run(
                            inputs=task_inputs,
                            run_context=run_context,
                            config_context=config_context,
                            ignore_position_hashes=ignore_position_hashes,
                        )
                    else:
                        failed_tasks.add(task)
                except Exception as e:
                    if config_context.fail_fast:
                        raise e
                    if config_context._swallow_exceptions:
                        exception = e
                        failed_tasks.add(task)
                    else:
                        raise e

        except Exception as e:
            if config_context.fail_fast:
                raise e
            exception = e

        return Result.init_from(
            subflow=flow,
            underlying=results,
            successful=(exception is None),
            task_values=results,
            exception=exception,
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import datetime as dt
import string
import sys
from typing import Any

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest
import sqlalchemy as sa
from packaging.version import Version
from pandas.core.dtypes.base import ExtensionDtype

import tests.util.tasks_library as m
from pydiverse.pipedag import (
    AUTO_VERSION,
    ConfigContext,
    Flow,
    Stage,
    StageLockContext,
    Table,
    materialize,
)
from pydiverse.pipedag.backend import SQLTableStore
from pydiverse.pipedag.backend.table.sql.hooks import PandasTableHook
from pydiverse.pipedag.backend.table.sql.sql import DISABLE_DIALECT_REGISTRATION
from pydiverse.pipedag.engine import dask

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import DATABASE_INSTANCES, with_instances
from tests.util.spy import spy_task
from tests.util.sql import get_config_with_table_store

# disable duckdb for now, since they have a bug in version 0.9.2 that needs fixing
pytestmark = [with_instances(tuple(set(DATABASE_INSTANCES) - {"duckdb"}))]
pd_version = Version(pd.__version__)


class Values:
    INT8 = [-128, 127]
    INT16 = [-32768, 32767]
    INT32 = [-2147483648, 2147483647]
    INT64 = [-9223372036854775808, 9223372036854775807]

    FLOAT32 = [3e38, 3.14159]
    FLOAT64 = [1e300, 3.14159]

    STR = ["", string.printable + string.whitespace]

    BOOLEAN = [True, False]

    DATE = [dt.date(1, 1, 1), dt.date(9999, 12, 31)]
    DATE_NS = [dt.date(1700, 1, 1), dt.date(2200, 1, 1)]
    TIME = [dt.time(0, 0, 0), dt.time(23, 59, 59)]
    DATETIME = [dt.datetime(1, 1, 1, 0, 0, 0), dt.datetime(9999, 12, 31, 23, 59, 59)]
    DATETIME_NS = [dt.datetime(1700, 1, 1), dt.datetime(2200, 1, 1)]


class NoneValues:
    INT = [0, None]
    FLOAT = [0.0, None]
    STR = ["", None]
    BOOLEAN = [False, None]

    DATE = [dt.date(1970, 1, 1), None]
    TIME = [dt.time(0, 0, 0), None]
    DATETIME = [dt.datetime(1970, 1, 1, 0, 0, 0), None]


def get_dialect_name():
    return ConfigContext.get().store.table_store.engine.dialect.name


class TestPandasTableHookNumpy:
    def test_basic(self):
        df = pd.DataFrame(
            {
                "int8": pd.array(Values.INT8, dtype="Int8"),
                "int16": pd.array(Values.INT16, dtype="Int16"),
                "int32": pd.array(Values.INT32, dtype="Int32"),
                "int64": pd.array(Values.INT64, dtype="Int64"),
                "float32": pd.array(Values.FLOAT32, dtype="Float32"),
                "float64": pd.array(Values.FLOAT32, dtype="Float64"),
                "str": pd.array(Values.STR, dtype="string"),
                "boolean": pd.array(Values.BOOLEAN, dtype="boolean"),
            }
        )

        @materialize()
        def numpy_input():
            return df

        @materialize(input_type=(pd.DataFrame, "numpy"))
        def assert_expected(in_df):
            pd.testing.assert_frame_equal(in_df, df, check_dtype=False)

            allowed_dtypes = set(df.dtypes)
            for dtype in in_df.dtypes:
                assert dtype in allowed_dtypes

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                assert_expected(t)

        assert f.run().successful

    def test_none(self):
        df = pd.DataFrame(
            {
                "int": pd.array(NoneValues.INT, dtype="Int64"),
                "float": pd.array(NoneValues.FLOAT, dtype="Float64"),
                "str": pd.array(NoneValues.STR, dtype="string"),
                "boolean": pd.array(NoneValues.BOOLEAN, dtype="boolean"),
            }
        )

        @materialize()
        def numpy_input():
            return df

        @materialize(input_type=(pd.DataFrame, "numpy"))
        def assert_expected(in_df):
            pd.testing.assert_frame_equal(in_df, df, check_dtype=False)

            allowed_dtypes = set(df.dtypes)
            for col, dtype in in_df.dtypes.items():
                if col == "boolean" and get_dialect_name() == "ibm_db_sa":
                    assert dtype == pd.Int16Dtype()
                else:
                    assert dtype in allowed_dtypes

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                assert_expected(t)

        assert f.run().successful

    @pytest.mark.skipif(pd.__version__ < "2.", reason="datetime64[us] requires pandas 2")
    def test_datetime(self):
        df = pd.DataFrame(
            {
                "date": Values.DATE,
                "date_ns": Values.DATE_NS,
                "date_none": NoneValues.DATE,
                "datetime": Values.DATETIME,
                "datetime_ns": Values.DATETIME_NS,
                "datetime_none": NoneValues.DATETIME,
            }
        ).astype(object)

        df_expected = pd.DataFrame(
            {
                "date": pd.array(Values.DATE, dtype="datetime64[s]"),
                "date_ns": pd.array(Values.DATE_NS, dtype="datetime64[s]"),
                "date_none": pd.array(NoneValues.DATE, dtype="datetime64[s]"),
                "datetime": pd.array(Values.DATETIME, dtype="datetime64[us]"),
                "datetime_ns": pd.array(Values.DATETIME_NS, dtype="datetime64[us]"),
                "datetime_none": pd.array(NoneValues.DATETIME, dtype="datetime64[us]"),
            }
        )

        @materialize()
        def numpy_input():
            datetime_dtype = sa.DateTime() if (get_dialect_name() != "mssql") else sa.dialects.mssql.DATETIME2()

            return Table(
                df,
                type_map={
                    "date": sa.Date(),
                    "date_ns": sa.Date(),
                    "date_none": sa.Date(),
                    "datetime": datetime_dtype,
                    "datetime_ns": datetime_dtype,
                    "datetime_none": datetime_dtype,
                },
            )

        @materialize(input_type=(pd.DataFrame, "numpy"))
        def assert_expected(in_df):
            pd.testing.assert_frame_equal(in_df, df_expected, check_dtype=False)

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                assert_expected(t)

        assert f.run().successful

    @pytest.mark.polars
    def test_datetime_polars(self):
        import polars as pl

        df = pd.DataFrame(
            {
                "date": Values.DATE,
                "date_ns": Values.DATE_NS,
                "date_none": NoneValues.DATE,
                "datetime": Values.DATETIME,
                "datetime_ns": Values.DATETIME_NS,
                "datetime_none": NoneValues.DATETIME,
            }
        ).astype(object)

        df_expected = pl.DataFrame(
            {
                "date": Values.DATE,
                "date_ns": Values.DATE_NS,
                "date_none": NoneValues.DATE,
                "datetime": Values.DATETIME,
                "datetime_ns": Values.DATETIME_NS,
                "datetime_none": NoneValues.DATETIME,
            }
        )

        @materialize()
        def numpy_input():
            datetime_dtype = sa.DateTime() if (get_dialect_name() != "mssql") else sa.dialects.mssql.DATETIME2()

            return Table(
                df,
                type_map={
                    "date": sa.Date(),
                    "date_ns": sa.Date(),
                    "date_none": sa.Date(),
                    "datetime": datetime_dtype,
                    "datetime_ns": datetime_dtype,
                    "datetime_none": datetime_dtype,
                },
            )

        @materialize(input_type=pl.DataFrame)
        def assert_expected(in_df):
            import polars.testing as pl_testing

            pl_testing.assert_frame_equal(in_df, df_expected, check_dtype=False)

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                assert_expected(t)

        assert f.run().successful


@pytest.mark.skipif(pd_version < Version("2.0"), reason="Requires pandas v2.0")
class TestPandasTableHookArrow:
    def test_basic(self):
        df = pd.DataFrame(
            {
                "int8": pd.array(Values.INT8, dtype="int8[pyarrow]"),
                "int16": pd.array(Values.INT16, dtype="int16[pyarrow]"),
                "int32": pd.array(Values.INT32, dtype="int32[pyarrow]"),
                "int64": pd.array(Values.INT64, dtype="int64[pyarrow]"),
                "float32": pd.array(Values.FLOAT32, dtype="float[pyarrow]"),
                "float64": pd.array(Values.FLOAT32, dtype="double[pyarrow]"),
                "str": pd.array(Values.STR, dtype=pd.ArrowDtype(pa.string())),
                "boolean": pd.array(Values.BOOLEAN, dtype="bool[pyarrow]"),
                "date": pd.array(Values.DATE, dtype=pd.ArrowDtype(pa.date32())),
                "time": pd.array(Values.TIME, dtype=pd.ArrowDtype(pa.time64("us"))),
                "datetime": pd.array(Values.DATETIME, dtype=pd.ArrowDtype(pa.timestamp("us"))),
            }
        )

        @materialize()
        def arrow_input():
            return df

        @materialize(input_type=(pd.DataFrame, "arrow"))
        def assert_expected(in_df):
            if ConfigContext.get().store.table_store.engine.dialect.name == "mssql":
                # these are unavoidable differences when using bcp for writing tables
                df["str"] = df["str"].replace("", pd.NA)
                df["str"] = (
                    df["str"]
                    .str.replace("\t", "\\t", regex=False)
                    .str.replace("\n", "\\n", regex=False)
                    .str.replace("\r", "\\r", regex=False)
                )
            pd.testing.assert_frame_equal(in_df, df, check_dtype=False)
            allowed_dtypes = set(df.dtypes)
            # Prefer StringDtype("pyarrow") over ArrowDtype(pa.string()) for now.
            # We need to check this choice with future versions of pandas/pyarrow.
            allowed_dtypes.remove(pd.ArrowDtype(pa.string()))
            allowed_dtypes.add(pd.StringDtype("pyarrow"))
            for col, dtype in in_df.dtypes.items():
                if col == "boolean" and get_dialect_name() == "ibm_db_sa":
                    assert dtype == pd.ArrowDtype(pa.int16())
                else:
                    assert dtype in allowed_dtypes

        with Flow() as f:
            with Stage("stage"):
                t = arrow_input()
                assert_expected(t)

        assert f.run().successful

    def test_none(self):
        df = pd.DataFrame(
            {
                "int": pd.array(NoneValues.INT, dtype="int64[pyarrow]"),
                "float": pd.array(NoneValues.FLOAT, dtype="double[pyarrow]"),
                "str": pd.array(NoneValues.STR, dtype=pd.ArrowDtype(pa.string())),
                "boolean": pd.array(NoneValues.BOOLEAN, dtype="bool[pyarrow]"),
                "date": pd.array(NoneValues.DATE, dtype=pd.ArrowDtype(pa.date32())),
                "time": pd.array(NoneValues.TIME, dtype=pd.ArrowDtype(pa.time64("us"))),
                "datetime": pd.array(NoneValues.DATETIME, dtype=pd.ArrowDtype(pa.timestamp("us"))),
            }
        )

        @materialize()
        def arrow_input():
            return df

        @materialize(input_type=(pd.DataFrame, "arrow"))
        def assert_expected(in_df):
            if ConfigContext.get().store.table_store.engine.dialect.name == "mssql":
                # these are unavoidable differences when using bcp for writing tables
                df["str"] = df["str"].replace("", pd.NA)
            pd.testing.assert_frame_equal(in_df, df, check_dtype=False)

        with Flow() as f:
            with Stage("stage"):
                t = arrow_input()
                assert_expected(t)

        assert f.run().successful


@pytest.mark.xfail(
    reason=("The string '\\N' can't be materialized using the COPY FROM STDIN WITH CSV technique."),
    strict=True,
)
@with_instances("postgres")
def test_pandas_table_hook_postgres_null_string():
    data = {"strNA": ["", None, "\\N"]}

    with Flow() as f:
        with Stage("stage_0"):
            t = m.pd_dataframe(data)

    with ConfigContext.get().evolve(swallow_exceptions=True), StageLockContext():
        result = f.run()
        df = result.get(t, as_type=pd.DataFrame)

    assert df["strNA"][0] == ""
    assert pd.isna(df["strNA"][1])
    assert df["strNA"].fillna("X")[2] == "\\N", "This is a known issue that the string '\\N' is considered NA"


@with_instances("postgres", "local_table_store")
@pytest.mark.skipif(dask.dask is None, reason="AUTO_VERSION for pandas requires dask")
@pytest.mark.skipif(
    sys.version_info >= (3, 13),
    reason="Pandas AUTO_VERSION ComputationTracer might be deprecated since it is hard to maintain",
)
class TestPandasAutoVersion:
    def test_smoke(self, mocker):
        should_swap_inputs = False
        global_df = pd.DataFrame({"global_col": [1, 0, 1, 0]})

        @materialize(input_type=pd.DataFrame, version=AUTO_VERSION, nout=2)
        def in_tables():
            in_table_1 = pd.DataFrame({"col": [1, 2, 3, 4]})
            in_table_2 = pd.DataFrame({"col": [4, 3, 2, 1]})

            if should_swap_inputs:
                return Table(in_table_2), Table(in_table_1)
            else:
                return Table(in_table_1), Table(in_table_2)

        @materialize(input_type=pd.DataFrame, version=AUTO_VERSION)
        def noop(df):
            return df, Table(df)

        @materialize(input_type=pd.DataFrame, version=AUTO_VERSION)
        def global_df_task(df):
            df["col"] *= global_df["global_col"]
            return df

        with Flow() as f:
            with Stage("stage"):
                in_tables_ = in_tables()
                noop_ = noop(in_tables_[0])
                global_df_task(noop_[0])

        f.run()
        in_tables_spy = spy_task(mocker, in_tables_)

        f.run()
        in_tables_spy.assert_called_once()

        should_swap_inputs = True
        f.run()
        in_tables_spy.assert_called(2)


@with_instances("postgres")
class TestPandasCustomHook:
    def test_custom_upload(self):
        class TestTableStore(ConfigContext.get().store.table_store.__class__):
            _dialect_name = DISABLE_DIALECT_REGISTRATION
            # this subclass is just to make sure hooks of other tests are not affected
            pass

        class TestTableStore2(TestTableStore):
            # this tests that the delegation to parent hooks works
            pass

        cfg = get_config_with_table_store(ConfigContext.get(), TestTableStore2)

        @TestTableStore.register_table(pd, replace_hooks=[PandasTableHook])
        class CustomPandasUploadTableHook(PandasTableHook):
            @classmethod
            def upload_table(
                cls,
                table: Table[pd.DataFrame],
                schema: str,
                dtypes: dict[str, sa.types.TypeEngine],
                store: SQLTableStore,
                early: bool,
            ):
                df = table.obj
                df["custom_upload"] = True
                super().upload_table(table, schema, dtypes, store, early)

        df = pd.DataFrame(
            {
                "int8": pd.array(Values.INT8, dtype="Int8"),
            }
        )

        @materialize()
        def numpy_input():
            return df

        @materialize(input_type=pd.DataFrame)
        def verify_custom(t):
            assert "custom_upload" in t.columns

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                verify_custom(t)

        assert f.run(config=cfg).successful

    def test_custom_download(self):
        class TestTableStore(ConfigContext.get().store.table_store.__class__):
            _dialect_name = DISABLE_DIALECT_REGISTRATION
            # this subclass is just to make sure hooks of other tests are not affected
            pass

        cfg = get_config_with_table_store(ConfigContext.get(), TestTableStore)

        @TestTableStore.register_table(pd, replace_hooks=[PandasTableHook])
        class CustomPandasDownloadTableHook(PandasTableHook):
            @classmethod
            def download_table(
                cls,
                query: Any,
                store: SQLTableStore,
                dtypes: dict[str, ExtensionDtype | np.dtype] | None = None,
            ) -> pd.DataFrame:
                df = super().download_table(query, store, dtypes)
                df["custom_download"] = True
                return df

        df = pd.DataFrame(
            {
                "int8": pd.array(Values.INT8, dtype="Int8"),
            }
        )

        @materialize()
        def numpy_input():
            return df

        @materialize(input_type=pd.DataFrame)
        def verify_custom(t):
            assert "custom_download" in t.columns

        with Flow() as f:
            with Stage("stage"):
                t = numpy_input()
                verify_custom(t)

        assert f.run(config=cfg).successful



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import tempfile

import sqlalchemy as sa

from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import Flow, Stage, Table, materialize
from pydiverse.pipedag.core.config import create_basic_pipedag_config


@materialize(lazy=True)
def lazy_task_1():
    return Table(sa.text("<invalid query>")).materialize()


def main():
    with tempfile.TemporaryDirectory() as temp_dir:
        cfg = create_basic_pipedag_config(
            f"duckdb:///{temp_dir}/db.duckdb",
            disable_stage_locking=True,  # This is special for duckdb
            # Attention: If uncommented, stage and task names might be sent to the
            #   following URL. You can self-host kroki if you like:
            #   https://docs.kroki.io/kroki/setup/install/
            #   You need to install optional dependency 'pydot' for any visualization
            #   URL to appear.
            # kroki_url="https://kroki.io",
        ).get("default")
        with cfg:
            with Flow() as f:
                with Stage("stage_1"):
                    lazy_task_1()

            # Run flow
            f.run()


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish
    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import sys
import warnings

import structlog

from pydiverse.common.util import requires
from pydiverse.pipedag import ExternalTableReference, Table
from pydiverse.pipedag.context import ConfigContext, RunContext
from pydiverse.pipedag.core import Result, Subflow, Task
from pydiverse.pipedag.engine.base import (
    OrchestrationEngine,
)

try:
    import dask
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    dask = None


@requires(dask, ImportError("DaskEngine requires 'dask' to be installed."))
class DaskEngine(OrchestrationEngine):
    """
    Execute a flow in parallel on a single machine using `dask <https://dask.org>`_.

    :param dask_compute_kwargs:
        Keyword arguments that get passed to :py:func:`dask.compute`.
        The main kwarg you might be interested in is ``num_workers``,
        which allows you to specify how many worker processes dask should spawn.
        By default, it spawns one worker per CPU core.
    """

    def __init__(self, **dask_compute_kwargs):
        self.dask_compute_kwargs = dict(
            traverse=True,
            optimize_graph=False,
            scheduler="processes",
            # Scheduler kwargs
            num_workers=None,
            chunksize=1,
        )

        self.dask_compute_kwargs.update(dask_compute_kwargs)

    def run(
        self,
        flow: Subflow,
        ignore_position_hashes: bool = False,
        inputs: dict[Task, ExternalTableReference] | None = None,
        **run_kwargs,
    ):
        inputs = inputs if inputs is not None else {}
        _ = ignore_position_hashes
        run_context = RunContext.get()
        config_context = ConfigContext.get()

        results = {}
        exception = None

        def bind_run(t: Task):
            structlog_config = structlog.get_config()
            structlog_context = structlog.contextvars.get_contextvars()

            def run(parent_futures, **kwargs):
                _ = parent_futures

                # TODO: Don't just assume a logger factory...
                structlog_config["logger_factory"] = structlog.PrintLoggerFactory(sys.stderr)
                structlog.configure(**structlog_config)

                with structlog.contextvars.bound_contextvars(**structlog_context):
                    return t.run(**kwargs)

            run.__name__ = t.name
            return dask.delayed(run, pure=False)

        for task in flow.get_tasks():
            task_inputs = {
                **{
                    in_id: results[in_t]
                    for in_id, in_t in task.input_tasks.items()
                    if in_t in results and in_t not in inputs
                },
                **{in_id: Table(inputs[in_t]) for in_id, in_t in task.input_tasks.items() if in_t in inputs},
            }

            results[task] = bind_run(task)(
                parent_futures=[results[parent] for parent in flow.get_parent_tasks(task)],
                inputs=task_inputs,
                run_context=run_context,
                config_context=config_context,
            )

        try:
            results = dask.compute(results, **self.dask_compute_kwargs)[0]
        except Exception as e:
            if config_context.fail_fast:
                raise e
            exception = e

        return Result.init_from(
            subflow=flow,
            underlying=results,
            successful=(exception is None),
            task_values=results,
            exception=exception,
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import sqlalchemy as sa

from pydiverse.pipedag import Flow, Stage, Table, materialize
from tests.fixtures.instances import (
    DATABASE_INSTANCES,
    ORCHESTRATION_INSTANCES,
    with_instances,
)


@materialize(nout=2, version="1.1")
def inputs():
    df_a = pd.DataFrame(
        {
            "a": [0, 1, 2, 4],
            "b": [9, 8, 7, 6],
        }
    )

    df_b = pd.DataFrame(
        {
            "a": [2, 1, 0, 1],
            "x": [1, 1, 2, 2],
        }
    )
    return Table(df_a, "dfA", primary_key=["a"]), Table(df_b, "dfB")


@materialize(input_type=pd.DataFrame, version="1.0")
def double_values(df: pd.DataFrame):
    df["a"] = df["a"] * 2
    return Table(df)


@materialize(input_type=sa.Table, lazy=True)
def join_on_a(left: sa.sql.expression.Alias, right: sa.sql.expression.Alias):
    return Table(left.select().join(right, left.c.a == right.c.a))


# noinspection PyTypeChecker
def get_flow():
    with Flow() as flow:
        with Stage("simple_flow_stage1"):
            a, b = inputs()
            a2 = double_values(a)

        with Stage("simple_flow_stage2"):
            b2 = double_values(b)
            joined = join_on_a(a2, b2)
            _ = joined
    return flow


@with_instances(DATABASE_INSTANCES, ORCHESTRATION_INSTANCES)
def test_simple_flow():
    flow = get_flow()
    result = flow.run()
    assert result.successful


if __name__ == "__main__":
    test_simple_flow()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import pytest

from pydiverse.pipedag import Flow, Stage, Table, materialize

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import DATABASE_INSTANCES, skip_instances, with_instances
from tests.util.tasks_library import assert_table_equal

pytestmark = [pytest.mark.ibis, with_instances(DATABASE_INSTANCES)]


try:
    import ibis
except ImportError:
    ibis = None


# connectorx and thus ibis have trouble with db2+ibm_db:// URLs and mssql
@skip_instances("ibm_db2", "mssql")
@pytest.mark.skipif(ibis is None, reason="ibis not installed")
def test_table_store():
    IbisTable = ibis.api.Table

    @materialize()
    def in_table():
        return Table(
            pd.DataFrame(
                {
                    "col": [0, 1, 2, 3],
                }
            )
        )

    @materialize()
    def expected_out_table():
        return Table(
            pd.DataFrame(
                {
                    "col": [0, 1, 2, 3],
                    "x": [1, 1, 1, 1],
                    "y": [2, 2, 2, 2],
                }
            )
        )

    @materialize(input_type=IbisTable)
    def noop(x):
        return Table(x)

    @materialize(lazy=True, input_type=IbisTable)
    def noop_lazy(x):
        return Table(x)

    @materialize(input_type=IbisTable)
    def add_column(x: IbisTable):
        return Table(x.mutate(x=ibis.literal(1)))

    @materialize(lazy=True, input_type=IbisTable)
    def add_column_lazy(x: IbisTable):
        return Table(x.mutate(y=ibis.literal(2)))

    with Flow() as f:
        with Stage("ibis"):
            table = in_table()
            table = noop(table)
            table = noop_lazy(table)
            table = add_column(table)
            table = add_column_lazy(table)

            expected = expected_out_table()
            _ = assert_table_equal(table, expected, check_dtype=False)

    assert f.run().successful



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.common.errors import DisposedError


class FlowError(Exception):
    """
    Exception raised when there is an issue with the flow definition.
    """


class StageError(Exception):
    """
    Exception raised when something is wrong with the stage.
    """


class GroupNodeError(Exception):
    """
    Exception raised when something is wrong with the stage.
    """


class CacheError(Exception):
    """
    Exception raised if something couldn't be retrieved from the cache.
    """


class LockError(Exception):
    """
    Exception raised if something goes wrong while locking, for example if
    a lock expires before it has been released.
    """


class DuplicateNameError(ValueError):
    """
    Exception raised if an object that is supposed to have a unique name doesn't.
    """


class IPCError(Exception):
    """
    Exception raised when inter process communication fails.
    """


class RemoteProcessError(IPCError):
    """
    Exception raised if an exception occurred in the remote IPC process.
    """


class HookCheckException(Exception):
    """
    Exception raised if a hook check fails. It is caught to surface as a
    normal task failure.
    """


class StoreIncompatibleException(Exception):
    """
    Exception raised if a store is incompatible for retrieving a table type.
    """


__all__ = [
    "FlowError",
    "StageError",
    "GroupNodeError",
    "CacheError",
    "LockError",
    "DuplicateNameError",
    "IPCError",
    "RemoteProcessError",
    "DisposedError",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import tempfile

import pandas as pd
import sqlalchemy as sa

from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import Schema, Table, materialize
from pydiverse.pipedag.core.config import create_basic_pipedag_config
from pydiverse.pipedag.materialize.debug import materialize_table


def main():
    with tempfile.TemporaryDirectory() as temp_dir:
        cfg = get_config(temp_dir)
        engine = cfg.store.table_store.engine

        # We can call tasks interactively which, however, will not do materialization
        # nor dematerialization. Thus we can only link tasks with same input_type:
        a, b = eager_inputs()
        pd.testing.assert_frame_equal(a, dfA)
        a2 = a.rename(columns={"a": "x"})  # we need a dataframe with a column named "x"
        eager = eager_task(a2, b)

        # Same also works for lazy tasks, however, SQLAlchemy expressions are a bit
        # harder to manage manually:
        lazy_1 = lazy_task_1()
        assert str(lazy_1.compile(engine, compile_kwargs={"literal_binds": True})) == "SELECT 1 AS x, 2 AS y"

        # We might want to create a custom debugging schema:
        schema = Schema(
            "any_schema_will_do",
            prefix=cfg.store.table_store.schema_prefix,
            suffix=cfg.store.table_store.schema_suffix,
        )
        with engine.connect() as conn:
            conn.execute(sa.text(f"CREATE SCHEMA IF NOT EXISTS {schema.get()}"))
            if sa.__version__ >= "2.0.0":
                conn.commit()
        # We can do this also a bit shorter and with more debugging output:
        cfg.store.table_store.execute(f"CREATE SCHEMA IF NOT EXISTS {schema.get()}")

        # Between lazy tasks we need to materialize.
        # However, we can give a configuration / database URL quite easily:
        tbl = Table(lazy_task_1(), name="task_1_out").materialize(cfg, schema)
        assert tbl.original.name == "task_1_out"
        assert tbl.original.schema == schema.get()

        # It is also possible to explicitly ask for materialize() to return dataframes
        assert Table(lazy_task_1()).materialize(cfg, schema, return_as_type=pd.DataFrame)["x"][0] == 1

        # ## now we can call the following flow interactively:
        # with cfg:
        #     with Flow() as f:
        #         with Stage("stage_1"):
        #             lazy_1 = lazy_task_1()
        #             a, b = eager_inputs()
        #
        #         with Stage("stage_2"):
        #             lazy_2 = lazy_task_2(lazy_1, b)
        #             lazy_3 = lazy_task_3(lazy_2)
        #             eager = eager_task(lazy_1, b)
        #
        #         with Stage("stage_3"):
        #             lazy_4 = lazy_task_4(lazy_2)
        #         _ = lazy_3, lazy_4, eager  # unused terminal output tables

        # stage_1:
        stage_1 = Schema("stage_1")
        cfg.store.table_store.execute(f"CREATE SCHEMA IF NOT EXISTS {stage_1.get()}")
        lazy_1 = lazy_task_1()
        a, b = eager_inputs()
        # materializations and dematerializations:
        lazy_1, lazy_1_df = Table(lazy_1).materialize(
            config_context=cfg, schema=stage_1, return_as_type=[None, pd.DataFrame]
        )
        a_df = a
        b_df = b
        b = Table(b).materialize(config_context=cfg, schema=stage_1)
        # stage_2:
        stage_2 = Schema("stage_2")
        cfg.store.table_store.execute(f"CREATE SCHEMA IF NOT EXISTS {stage_2.get()}")
        lazy_2 = lazy_task_2(lazy_1, b)
        lazy_2 = Table(lazy_2, name="task_2_out").materialize(config_context=cfg, schema=stage_2)
        lazy_3 = lazy_task_3(lazy_2)
        eager = eager_task(lazy_1_df, b_df)
        # stage_3:
        stage_3 = Schema("stage_3")
        cfg.store.table_store.execute(f"CREATE SCHEMA IF NOT EXISTS {stage_3.get()}")
        lazy_4 = lazy_task_4(lazy_2, cfg, stage_3)

        # testing:
        pd.testing.assert_frame_equal(a_df, dfA)
        assert (
            str(lazy_3.compile(engine, compile_kwargs={"literal_binds": True}))
            == 'SELECT * FROM "stage_2"."task_2_out"'
        )
        assert sorted(eager.columns) == ["a", "x", "y"]
        assert (
            str(lazy_4.compile(engine, compile_kwargs={"literal_binds": True}))
            == '\n        SELECT * FROM "stage_2"."task_2_out" as input1\n        '
            'LEFT JOIN "stage_3"."_sub" as sub ON input1.a = sub.a\n    '
        )

        # it is also possible to run code inside tasks in exactly the same way:
        # lazy_4 = lazy_task_4(lazy_2, cfg, stage_3):
        input1 = lazy_2
        subquery = f"""
            SELECT input1.a, sum(input1.x5) as x_sum FROM {ref(input1)} as input1
            GROUP BY a
        """
        sub_ref = Table(sa.text(subquery), name="_sub").materialize(cfg, stage_3)
        query = f"""
            SELECT * FROM {ref(input1)} as input1
            LEFT JOIN {ref(sub_ref)} as sub ON input1.a = sub.a
        """
        lazy_4 = Table(sa.text(query), name="enriched_aggregation").materialize()
        assert (
            str(lazy_4.compile(engine, compile_kwargs={"literal_binds": True}))
            == '\n            SELECT * FROM "stage_2"."task_2_out" as input1\n         '
            '   LEFT JOIN "stage_3"."_sub" as sub ON input1.a = sub.a\n        '
        )

        # eager = eager_task(lazy_1_df, b_df):
        tbl1, tbl2 = lazy_1_df, b_df
        eager = Table(tbl1.merge(tbl2, on="x")).materialize()
        assert sorted(eager.columns) == ["a", "x", "y"]

        # there is a more low level function to debug materialize tables with a few more
        # options exposed:
        materialize_table(Table(lazy_task_1(), name="task_1_out"), cfg, stage_1, debug_suffix="_debug")


@materialize(lazy=True)
def lazy_task_1():
    return Table(
        sa.select(
            sa.literal(1).label("x"),
            sa.literal(2).label("y"),
        ),
        name="task_1_out",
    ).materialize()


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_2(input1: sa.sql.expression.Alias, input2: sa.sql.expression.Alias):
    query = sa.select(
        (input1.c.x * 5).label("x5"),
        input2.c.a,
    ).select_from(input1.outerjoin(input2, input2.c.x == input1.c.x))

    return Table(query, name="task_2_out", primary_key=["a"]).materialize()


def ref(tbl: sa.sql.expression.Alias):
    return f'"{tbl.original.schema}"."{tbl.original.name}"'


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_3(input1: sa.sql.expression.Alias):
    return Table(sa.text(f"SELECT * FROM {ref(input1)}")).materialize()


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_4(input1: sa.sql.expression.Alias, config_context=None, schema=None):
    # imperatively materialize a subquery
    subquery = f"""
        SELECT input1.a, sum(input1.x5) as x_sum FROM {ref(input1)} as input1
        GROUP BY a
    """
    sub_ref = Table(sa.text(subquery), name="_sub").materialize(config_context, schema)
    query = f"""
        SELECT * FROM {ref(input1)} as input1
        LEFT JOIN {ref(sub_ref)} as sub ON input1.a = sub.a
    """
    return Table(sa.text(query), name="enriched_aggregation").materialize()


dfA = pd.DataFrame(
    {
        "a": [0, 1, 2, 4],
        "b": [9, 8, 7, 6],
    }
)


@materialize(nout=2, version="1.0.0")
def eager_inputs():
    dfB = pd.DataFrame(
        {
            "a": [2, 1, 0, 1],
            "x": [1, 1, 2, 2],
        }
    )
    return Table(dfA, "dfA").materialize(), Table(dfB, "dfB_%%").materialize()


@materialize(version="1.0.0", input_type=pd.DataFrame)
def eager_task(tbl1: pd.DataFrame, tbl2: pd.DataFrame):
    return Table(tbl1.merge(tbl2, on="x")).materialize()


def get_config(temp_dir, filename: str = "db"):
    cfg = create_basic_pipedag_config(
        f"duckdb:///{temp_dir}/{filename}.duckdb",
        disable_stage_locking=True,  # This is special for duckdb
        # Attention: If uncommented, stage and task names might be sent to the
        #   following URL. You can self-host kroki if you like:
        #   https://docs.kroki.io/kroki/setup/install/
        #   You need to install optional dependency 'pydot' for any visualization
        #   URL to appear.
        # kroki_url="https://kroki.io",
    ).get("default")
    return cfg


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish
    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Callable

import structlog

from pydiverse.common.util import Disposable, deep_map
from pydiverse.common.util.hashing import stable_hash
from pydiverse.pipedag import Blob, Schema, Stage, Table
from pydiverse.pipedag._typing import Materializable, T
from pydiverse.pipedag.container import RawSql, attach_annotation
from pydiverse.pipedag.context import ConfigContext, RunContext, TaskContext
from pydiverse.pipedag.context.run_context import StageState
from pydiverse.pipedag.core.config import PipedagConfig
from pydiverse.pipedag.core.task import Task, TaskGetItem
from pydiverse.pipedag.errors import (
    CacheError,
    DuplicateNameError,
    HookCheckException,
    StageError,
    StoreIncompatibleException,
)
from pydiverse.pipedag.materialize.cache import TaskCacheInfo, lazy_table_cache_key
from pydiverse.pipedag.materialize.materializing_task import MaterializingTask
from pydiverse.pipedag.materialize.metadata import (
    LazyTableMetadata,
    RawSqlMetadata,
    TaskMetadata,
)
from pydiverse.pipedag.materialize.table_hook_base import TableHookResolver


class BaseTableStore(TableHookResolver, Disposable):
    """Table store base class

    The table store is responsible for storing and retrieving various types
    of tabular data. Additionally, it also has to manage all task metadata,
    This includes storing it, but also cleaning up stale metadata.

    A store must use a table's name (`table.name`) and stage (`table.stage`)
    as the primary keys for storing and retrieving it. This means that
    two different `Table` objects can be used to store and retrieve the same
    data as long as they have the same name and stage.

    The same is also true for the task metadata where the task `stage`,
    `version` and `cache_key` act as the primary keys (those values are
    stored both in the task object and the metadata object).

    To implement the stage transaction and commit mechanism, a technique
    called schema swapping is used:

    All outputs from materializing tasks get materialized into a temporary
    empty schema (`stage.transaction_name`) and only if all tasks have
    finished running *successfully* you swap the 'base schema' (original stage,
    or cache) with the 'transaction schema'. This is usually done by renaming
    them.

    """

    def __init__(self):
        self.logger = structlog.get_logger(logger_name=type(self).__name__)
        self.local_table_cache: BaseTableCache | None = None

    def setup(self):
        """Setup function

        This function gets called at the beginning of a flow run.
        Unlike the __init__ method, a lock is acquired before
        the setup method gets called to prevent race conditions.
        """

    # Stage

    @abstractmethod
    def init_stage(self, stage: Stage):
        """Initialize a stage transaction

        When working with schema swapping:

        Ensures that the base schema exists (but doesn't clear it) and that
        the transaction schema exists and is empty.
        """

    @abstractmethod
    def commit_stage(self, stage: Stage):
        """Commit the stage

        When using schema swapping:

        After the schema swap the contents of the base schema should be in the
        transaction schema, and the contents of the transaction schema in
        the base schema.

        Additionally, the metadata associated with the transaction schema should
        replace the metadata of the base schema. The latter can be discarded.
        """

    # Materialize

    def store_table(self, table: Table, task: MaterializingTask | None):
        if self.local_table_cache:
            self.local_table_cache.store_table(table, task)
        super().store_table(table, task)

    def execute_raw_sql(self, raw_sql: RawSql):
        """Executed raw SQL statements in the associated transaction stage

        This method is overridden by actual table stores that can handle raw SQL.
        """

        raise NotImplementedError("This table store does not support executing raw sql statements")

    def store_table_lazy(
        self,
        table: Table,
        task: MaterializingTask,
        task_cache_info: TaskCacheInfo,
    ):
        """Lazily stores a table in the associated commit stage

        The same as `store_table()`, with the difference being that if the
        table object represents a lazy table / query, the store first checks
        if the same query with the same input (based on `table.cache_key`)
        has already been executed before. If yes, instead of evaluating
        the query, it just copies the previous result to the commit stage.

        Used when `lazy = True` is set for a materializing task.
        """
        config_context = ConfigContext.get()
        try:
            hook = self.get_m_table_hook(table)
            query_str = hook.lazy_query_str(self, table.obj)
        except TypeError:
            self.logger.warning(
                f"The output table {table.name} given by a"
                f" {repr(type(table.obj))} of the lazy task {task.name} does"
                " not provide a query string. Lazy evaluation is not"
                " possible. Assuming that the table is not cache valid."
            )
            # Assign random query string to ensure that task is not cache valid
            query_str = uuid.uuid4().hex

        if table.assumed_dependencies is None:
            query_hash = stable_hash("LAZY-TABLE", query_str)
        else:
            # include assumed dependencies in query hash for imperative materialize
            dependencies = config_context.store.json_encoder.encode(table.assumed_dependencies)
            query_hash = stable_hash("LAZY-TABLE", query_str, dependencies)

        # Store the table
        try:
            if task_cache_info.force_task_execution:
                self.logger.info(
                    "Forced task execution due to config",
                    cache_validation=config_context.cache_validation,
                )
                raise CacheError("Forced task execution")
            # Try retrieving the table from the cache and then copying it
            # to the transaction stage
            metadata = self.retrieve_lazy_table_metadata(query_hash, task_cache_info.cache_key, table.stage)
            RunContext.get().trace_hook.query_cache_status(
                task,
                table,
                task_cache_info,
                query_hash,
                query_str,
                cache_metadata=metadata,
            )
            RunContext.get().trace_hook.cache_init_transfer(task, table)
            self.copy_lazy_table_to_transaction(metadata, table)
            self.logger.info(f"Lazy cache of table '{table.name}' found")
        except CacheError as e:
            # Either not found in cache, or copying failed
            # -> Store using default method
            self.logger.warning("Cache miss", table=table.name, stage=table.stage.name, cause=str(e))
            TaskContext.get().is_cache_valid = False
            RunContext.get().trace_hook.query_cache_status(
                task, table, task_cache_info, query_hash, query_str, cache_valid=False
            )
            if task_cache_info.assert_no_materialization:
                raise AssertionError(
                    "cache_validation.mode=ASSERT_NO_FRESH_INPUT is a "
                    "protection mechanism to prevent execution of "
                    "source tasks to keep pipeline input stable. However,"
                    "this table was still about to be materialized: "
                    f"{table.stage.name}.{table.name}"
                ) from None
            self.store_table(table, task)
            RunContext.get().trace_hook.query_complete(task, table)

        # Store table metadata
        self.store_lazy_table_metadata(
            LazyTableMetadata(
                name=table.name,
                stage=table.stage.name,
                query_hash=query_hash,
                task_hash=task_cache_info.cache_key,
            )
        )

        # At this point we MUST also update the cache info, so that any downstream
        # tasks get invalidated if the sql query string changed.
        table.cache_key = lazy_table_cache_key(task_cache_info.cache_key, query_hash)

    def store_raw_sql(self, raw_sql: RawSql, task: MaterializingTask, task_cache_info: TaskCacheInfo):
        """Lazily stores a table in the associated commit stage

        The same as `store_table()`, with the difference being that the store first
        checks if the same query with the same input (based on `raw_sql.cache_key`)
        has already been executed before. If yes, instead of evaluating
        the query, it just copies the previous result to the commit stage.
        """

        # hacky way to canonicalize query (despite __tmp/__even/__odd suffixes
        # and alias resolution)
        import re

        query_str = raw_sql.sql
        query_str = re.sub(r'["\[\]]', "", query_str)
        query_str = re.sub(r'(__tmp|__even|__odd)(?=[ \t\n.;"]|$)', "", query_str.lower())

        query_hash = stable_hash("RAW-SQL", query_str)

        # Store raw sql
        try:
            if task_cache_info.force_task_execution:
                config_context = ConfigContext.get()
                self.logger.info(
                    "Forced task execution due to config",
                    cache_validation=config_context.cache_validation,
                )
                raise CacheError("Forced task execution")
            # Try retrieving the table from the cache and then copying it
            # to the transaction stage
            metadata = self.retrieve_raw_sql_metadata(query_hash, task_cache_info.cache_key, raw_sql.stage)
            self.copy_raw_sql_tables_to_transaction(metadata, raw_sql.stage)
            self.logger.info(f"Lazy cache of stage '{raw_sql.stage}' found")

            prev_objects = metadata.prev_objects
            new_objects = metadata.new_objects
        except CacheError as e:
            # Either not found in cache, or copying failed
            # -> Store using default method
            self.logger.warning("Cache miss for raw-SQL", cause=str(e))

            TaskContext.get().is_cache_valid = False
            RunContext.get().set_stage_has_changed(task.stage)

            if task_cache_info.assert_no_materialization:
                raise AssertionError(
                    "cache_validation.mode=ASSERT_NO_FRESH_INPUT is a "
                    "protection mechanism to prevent execution of "
                    "source tasks to keep pipeline input stable. However,"
                    f"this raw SQL script was still about to be executed: {raw_sql}"
                ) from None

            prev_objects = self.get_objects_in_stage(raw_sql.stage)
            self.execute_raw_sql(raw_sql)
            post_objects = self.get_objects_in_stage(raw_sql.stage)

            # Object names must be sorted to ensure that we can identify the task
            # again in the future even if the objects get returned in a different order.
            prev_objects = sorted(prev_objects)

            prev_objects_set = set(prev_objects)
            new_objects = [o for o in post_objects if o not in prev_objects_set]

        # Store metadata
        # Attention: Raw SQL statements may only be executed sequentially within
        #            stage for store.get_objects_in_stage to work
        self.store_raw_sql_metadata(
            RawSqlMetadata(
                prev_objects=prev_objects,
                new_objects=new_objects,
                stage=raw_sql.stage.name,
                query_hash=query_hash,
                task_hash=task_cache_info.cache_key,
            )
        )

        # At this point we MUST also update the cache info, so that any downstream
        # tasks get invalidated if the sql query string changed.
        raw_sql.cache_key = lazy_table_cache_key(task_cache_info.cache_key, query_hash)

        # Store new_objects as part of raw_sql.
        all_table_names = set(self.get_table_objects_in_stage(raw_sql.stage))
        raw_sql.table_names = sorted(o for o in new_objects if o in all_table_names)

    @abstractmethod
    def copy_table_to_transaction(self, table: Table):
        """Copy a table from the base stage to the transaction stage

        This operation MUST not remove the table from the base stage store
        or modify it in any way.

        :raises CacheError: if the table can't be found in the cache
        """

    @abstractmethod
    def copy_lazy_table_to_transaction(self, metadata: LazyTableMetadata, table: Table):
        """Copy the lazy table identified by the metadata to the transaction stage of
        table.

        This operation MUST not remove the table from the base stage or modify
        it in any way.

        :raises CacheError: if the lazy table can't be found
        """

    def copy_raw_sql_tables_to_transaction(self, metadata: RawSqlMetadata, stage: Stage):
        """Copy all tables identified by the metadata as generated by raw
        SQL statements to the transaction stage.

        This operation MUST not remove the table from the base stage or modify
        it in any way.

        :raises CacheError: if the lazy table can't be found
        """
        raise NotImplementedError("This table store does not support executing raw sql statements")

    @abstractmethod
    def delete_table_from_transaction(self, table: Table, *, schema: Schema | None = None):
        """Delete a table from the transaction

        If the table doesn't exist in the transaction stage, fail silently.
        """

    def retrieve_table_obj(
        self,
        table: Table,
        as_type: type[T] | None,
        for_auto_versioning: bool = False,
    ) -> T:
        if as_type is None:
            # Simply return enough information that a user could dematerialize the table
            # or perform it with some other library.
            # hint: `schema = table_store.get_schema(table.stage.current_name).get()`
            return table.name, table.stage.current_name

        if for_auto_versioning:
            return super().retrieve_table_obj(table, as_type, for_auto_versioning)

        if self.local_table_cache:
            obj = self.local_table_cache.retrieve_table_obj(table, as_type)
            if obj is not None:
                return obj

        obj = super().retrieve_table_obj(table, as_type)

        if self.local_table_cache:
            t = table.copy_without_obj()
            t.obj = obj
            self.local_table_cache.store_input(t, task=None)

        return obj

    # Metadata

    @abstractmethod
    def store_task_metadata(self, metadata: TaskMetadata, stage: Stage):
        """Stores the metadata of a task

        The metadata must always be stored in such a way that it is
        associated with the transaction. Only after a stage has been
        committed, should it be associated with the base stage / cache.
        """

    @abstractmethod
    def retrieve_task_metadata(self, task: MaterializingTask, input_hash: str, cache_fn_hash: str) -> TaskMetadata:
        """Retrieve a task's metadata from the store

        :raises CacheError: if no metadata for this task can be found.
        """

    @abstractmethod
    def retrieve_all_task_metadata(
        self, task: MaterializingTask, ignore_position_hashes: bool = False
    ) -> list[TaskMetadata]:
        """Retrieves all metadata objects associated with a task from the store

        As long as a metadata entry has the same task and stage name, as well
        as the same position hash as the `task` object, it should get returned.
        """

    # Lazy Table Metadata

    @abstractmethod
    def store_lazy_table_metadata(self, metadata: LazyTableMetadata):
        """Stores the metadata of a lazy table

        The metadata must always be stored in such a way that it is
        associated with the transaction. Only after a stage has been
        committed, should it be associated with the base stage / cache.
        """

    @abstractmethod
    def retrieve_lazy_table_metadata(self, query_hash: str, task_hash: str, stage: Stage) -> LazyTableMetadata:
        """Retrieve a lazy table's metadata from the store

        :param query_hash: A hash of the query that produced this lazy table
        :param task_hash: The hash of the task for which we want to retrieve this
            metadata. This can be used to retrieve the lazy table metadata produced
            by the same task in a previous run, if the current task is still cache
            valid.
        :param stage: The stage in which this lazy table should be.
        :return: The metadata.

        :raises CacheError: if not metadata that matches the provided inputs was found.
        """

    def store_raw_sql_metadata(self, metadata: RawSqlMetadata):
        """Stores the metadata of raw SQL statements

        The metadata must always be stored in such a way that it is
        associated with the transaction. Only after a stage has been
        committed, should it be associated with the base stage / cache.
        """
        raise NotImplementedError("This table store does not support executing raw sql statements")

    def retrieve_raw_sql_metadata(self, query_hash: str, task_hash: str, stage: Stage) -> RawSqlMetadata:
        """Retrieve raw SQL metadata from the store

        :param query_hash: A hash of the query that produced this raw sql object
        :param task_hash: The hash of the task for which we want to retrieve this
            metadata. This can be used to retrieve the raw sql metadata produced
            by the same task in a previous run, if the current task is still cache
            valid.
        :param stage: The stage associated with the raw sql object.
        :return: The metadata.

        :raises CacheError: if not metadata that matches the provided inputs was found.
        """
        raise NotImplementedError("This table store does not support executing raw sql statements")

    # Utility

    @abstractmethod
    def get_objects_in_stage(self, stage: Stage) -> list[str]:
        """
        List all objects that are in the current stage.

        This may include tables but also other database objects like views, stored
        procedures, functions etc. This function is used to calculate a diff on the
        table store to determine which objects were produced (or could have been used
        to produce those objects) when executing RawSQL.

        :param stage: the stage
        :return: list of object names in the stage at the current point in time.
        """

    @abstractmethod
    def get_table_objects_in_stage(self, stage: Stage, include_views=True) -> list[str]:
        """
        List all table-like objects that are in the current stage.

        :param stage: the stage
        :return: list of table-like object names in the stage at
            the current point in time.
        """


class BaseTableCache(ABC, TableHookResolver, Disposable):
    def __init__(
        self,
        store_input: bool = True,
        store_output: bool = False,
        use_stored_input_as_cache: bool = True,
    ):
        super().__init__()

        self.logger = structlog.get_logger(logger_name=type(self).__name__)

        self.should_store_input = store_input
        self.should_store_output = store_output
        self.should_use_stored_input_as_cache = use_stored_input_as_cache

    def setup(self):
        """Setup function

        This function gets called at the beginning of a flow run.
        Unlike the __init__ method, a lock is acquired before
        the setup method gets called to prevent race conditions.
        """

    def init_stage(self, stage: Stage):
        """Initialize a stage

        Gets called before any table is attempted to be stored in the stage.
        """

    @abstractmethod
    def clear_cache(self, stage: Stage):
        """Delete the cache for a specific stage"""

    def store_table(self, table: Table, task: MaterializingTask):
        if self.should_store_output:
            return self._store_table(table, task)

    def store_input(self, table: Table, task: MaterializingTask):
        if self.should_store_input:
            return self._store_table(table, task)

    def _store_table(self, table: Table, task: MaterializingTask | None) -> bool:
        """
        :return: bool flag indicating if storing was successful
        """
        try:
            hook = self.get_m_table_hook(table)
        except TypeError:
            return False

        if not RunContext.get().should_store_table_in_cache(table):
            # Prevent multiple tasks writing at the same time
            return False

        try:
            hook.materialize(self, table, table.stage.transaction_name)
        except TypeError:
            return False
        return True

    def retrieve_table_obj(
        self,
        table: Table,
        as_type: type[T],
        for_auto_versioning: bool = False,
    ) -> T:
        assert not for_auto_versioning

        if not self.should_use_stored_input_as_cache:
            return None
        if not self._has_table(table, as_type):
            return None
        return self._retrieve_table_obj(table, as_type)

    def _retrieve_table_obj(self, table: Table, as_type: type[T]) -> T:
        try:
            hook = self.get_r_table_hook(as_type)
            obj = hook.retrieve(self, table, table.stage.name, as_type)
            self.logger.info("Retrieved table from local table cache", table=table)
            return obj
        except StoreIncompatibleException:
            # This is expected for example when ParquetTableCache is asked to retrieve a SQL reference
            return None
        except Exception as e:
            self.logger.warning(
                "Failed to retrieve table from local table cache",
                table=table,
                cause=str(e),
            )
            return None

    @abstractmethod
    def _has_table(self, table: Table, as_type: type) -> bool:
        """Check if the given table is in the cache"""


class BaseBlobStore(Disposable, ABC):
    """Blob store base class

    A blob (binary large object) store is responsible for storing arbitrary
    python objects. This can, for example, be done by serializing them using
    the python ``pickle`` module.

    A store must use a blob's name (``Blob.name``) and stage (``Blob.stage``)
    as the primary keys for storing and retrieving blobs. This means that
    two different ``Blob`` objects can be used to store and retrieve the same
    data as long as they have the same name and stage.
    """

    @abstractmethod
    def init_stage(self, stage: Stage):
        """Initialize a stage and start a transaction"""

    @abstractmethod
    def commit_stage(self, stage: Stage):
        """Commit the stage transaction

        Replace the blobs of the base stage with the blobs in the transaction.
        """

    @abstractmethod
    def store_blob(self, blob: Blob):
        """Stores a blob in the associated stage transaction"""

    @abstractmethod
    def copy_blob_to_transaction(self, blob: Blob):
        """Copy a blob from the base stage to the transaction

        This operation MUST not remove the blob from the base stage or modify
        it in any way.
        """

    @abstractmethod
    def delete_blob_from_transaction(self, blob: Blob):
        """Delete a blob from the transaction

        If the blob doesn't exist in the transaction, fail silently.
        """

    @abstractmethod
    def retrieve_blob(self, blob: Blob) -> Any:
        """Loads a blob from the store

        Retrieves the stored python object from the store and returns it.
        If the stage hasn't yet been committed, the blob must be retrieved
        from the transaction, else it must be retrieved from the committed
        stage.
        """


class PipeDAGStore(Disposable):
    """Main storage interface for materializing tasks

    Depending on the use case, the store can be configured using different
    backends for storing tables, blobs and managing locks.

    Other than initializing the global `PipeDAGStore` object, the user
    should never have to interact with it. It only serves as a coordinator
    between the different backends, the stages and the materializing tasks.
    """

    def __init__(
        self,
        table: BaseTableStore,
        blob: BaseBlobStore,
        local_table_cache: BaseTableCache | None,
    ):
        self.table_store = table
        self.blob_store = blob
        self.local_table_cache = local_table_cache
        self.table_store.local_table_cache = local_table_cache

        self.logger = structlog.get_logger()

        from pydiverse.pipedag.util.json import PipedagJSONDecoder, PipedagJSONEncoder

        self.json_encoder = PipedagJSONEncoder()
        self.json_decoder = PipedagJSONDecoder()

    def dispose(self):
        """
        Clean up and close all open resources.

        Don't use the store object any more after disposal!
        """
        self.table_store.dispose()
        self.blob_store.dispose()
        if self.local_table_cache:
            self.local_table_cache.dispose()
        super().dispose()

    # ### Stage ### #

    def init_stage(self, stage: Stage):
        """Initializes the stage in all backends

        This function also acquires a lock on the given stage to prevent
        other flows from modifying the same stage at the same time. Only
        once all tasks that depend on this stage have been executed, will
        this lock be released.

        Don't use this function directly. Instead, use `ensure_stage_is_ready`
        to prevent unnecessary locking.
        """

        RunContext.get().validate_stage_lock(stage)
        self.table_store.init_stage(stage)
        self.blob_store.init_stage(stage)
        if self.local_table_cache:
            self.local_table_cache.init_stage(stage)

        # Once stage reference counter hits 0 (this means all tasks that
        # have any task contained in the stage as an upstream dependency),
        # the lock gets automatically released by the RunContextServer

    def ensure_stage_is_ready(self, stage: Stage):
        """Initializes a stage if it hasn't been created yet

        This allows the creation of stages in a lazy way. This ensures
        that a stage only gets locked right before it is needed.
        """
        with RunContext.get().init_stage(stage) as should_continue:
            if not should_continue:
                return
            self.init_stage(stage)

    def commit_stage(self, stage: Stage):
        """Commit the stage"""
        ctx = RunContext.get()
        with ctx.commit_stage(stage) as should_continue:
            if not should_continue:
                raise StageError
            ctx.validate_stage_lock(stage)
            self.table_store.commit_stage(stage)
            self.blob_store.commit_stage(stage)

    # ### Materialization ### #

    def dematerialize_item(
        self,
        item: Table | RawSql | Blob | Any,
        as_type: type[T] | None,
        ctx: RunContext | None = None,
        for_auto_versioning: bool = False,
    ):
        if ctx is None:
            ctx = RunContext.get()

        if isinstance(item, Table):
            # item.stage can be None for ExternalTableReference
            if item.stage is not None:
                ctx.validate_stage_lock(item.stage)
            obj = self.table_store.retrieve_table_obj(
                item,
                as_type=as_type,
                for_auto_versioning=for_auto_versioning,
            )
            return obj
        elif isinstance(item, RawSql):
            ctx.validate_stage_lock(item.stage)

            loaded_tables = {}
            for table_name in item:
                table = item[table_name]
                obj = self.table_store.retrieve_table_obj(
                    table,
                    as_type=as_type,
                    for_auto_versioning=for_auto_versioning,
                )
                loaded_tables[table_name] = obj

            new_raw_sql = item.copy_without_obj()
            new_raw_sql.loaded_tables = loaded_tables
            return new_raw_sql
        elif isinstance(item, Blob):
            ctx.validate_stage_lock(item.stage)
            return self.blob_store.retrieve_blob(item)
        return item

    def dematerialize_task_inputs(
        self,
        task: MaterializingTask,
        args: tuple[Materializable],
        kwargs: dict[str, Materializable],
        *,
        for_auto_versioning: bool = False,
    ) -> tuple[tuple, dict]:
        """Loads the inputs for a task from the storage backends

        Traverses the function arguments and replaces all `Table` and
        `Blob` objects with the associated objects stored in the backend.

        :param task: The task for which the arguments should be dematerialized
        :param args: The positional arguments
        :param kwargs: The keyword arguments
        :param for_auto_versioning: If the task inputs should be retrieved
            for use with auto versioning.
        :return: A tuple with the dematerialized args and kwargs
        """

        ctx = RunContext.get()

        def dematerialize_mapper(x):
            ret = self.dematerialize_item(
                x,
                as_type=task.input_type,
                ctx=ctx,
                for_auto_versioning=for_auto_versioning,
            )
            if task.add_input_source and isinstance(x, (Table, Blob, RawSql)):
                return ret, x
            return ret

        d_args = deep_map(args, dematerialize_mapper)
        d_kwargs = deep_map(kwargs, dematerialize_mapper)

        return d_args, d_kwargs

    def materialize_task(
        self,
        task: MaterializingTask,
        task_cache_info: TaskCacheInfo,
        value: Materializable,
        disable_task_finalization=False,
    ) -> Materializable:
        """Stores the output of a task in the backend

        Traverses the output produced by a task, adds missing metadata,
        materializes all `Table` and `Blob` objects and returns a new
        output object with the required metadata to allow dematerialization.

        :param task: The task instance which produced `value`. Must have
            the correct `cache_key` attribute set.
        :param task_cache_info: Task cache information.
        :param value: The output of the task. Must be materializable; this
            means it can only contain the following object types:
            `dict`, `list`, `tuple`,
            `int`, `float`, `str`, `bool`, `None`,
            and PipeDAG's `Table` and `Blob` type.
        :param disable_task_finalization: True for imperative materialization
            when task is not finished, yet.
        :return: A copy of `value` with additional metadata
        """

        stage = task.stage
        ctx = RunContext.get()

        if (state := ctx.get_stage_state(stage)) != StageState.READY:
            raise StageError(f"Can't materialize because stage '{stage.name}' is not ready (state: {state}).")

        value, tables, raw_sqls, blobs = self.prepare_task_output_for_materialization(task, task_cache_info, value)

        def store_metadata():
            """
            Metadata must be generated after everything else has been materialized,
            because during materialization the cache_key of the different objects
            can get changed.
            """
            if not disable_task_finalization:
                output_json = self.json_encode(value)
                metadata = TaskMetadata(
                    name=task.name,
                    stage=task.stage.name,
                    version=task.version,
                    timestamp=datetime.now(),
                    run_id=ctx.run_id,
                    position_hash=task.position_hash,
                    input_hash=task_cache_info.input_hash,
                    cache_fn_hash=task_cache_info.cache_fn_hash,
                    output_json=output_json,
                )
                self.table_store.store_task_metadata(metadata, stage)

        def store_table(table: Table):
            if task.lazy:
                self.table_store.store_table_lazy(table, task, task_cache_info)
            else:
                self.table_store.store_table(table, task)

        # Materialize
        self._check_names(task, tables, blobs)
        check_failures = self._store_task_transaction(
            task,
            tables,
            raw_sqls,
            blobs,
            store_table,
            lambda raw_sql: self.table_store.store_raw_sql(raw_sql, task, task_cache_info),
            self.blob_store.store_blob,
            store_metadata,
        )
        if len(check_failures) > 0:
            self.logger.error(
                f"{len(check_failures)} output validation checks failed",
                failures=check_failures,
            )
            raise check_failures[0][1]

        return value

    def prepare_task_output_for_materialization(
        self,
        task: MaterializingTask,
        task_cache_info: TaskCacheInfo,
        value: Materializable,
    ) -> tuple[Materializable, list[Table], list[RawSql], list[Blob]]:
        tables = []
        raw_sqls = []
        blobs = []

        config = ConfigContext.get()
        auto_suffix_counter = task_cache_info.imperative_materialization_state.auto_suffix_counter

        def preparation_mutator(x):
            # Automatically convert an object to a table / blob if its
            # type is inside either `config.auto_table` or `.auto_blob`.
            if isinstance(x, config.auto_table):
                try:
                    hook = self.table_store.get_m_table_hook(Table(x))
                    x = hook.auto_table(x)
                except TypeError:
                    x = Table(x)
            if isinstance(x, config.auto_blob):
                x = Blob(x)

            if isinstance(x, PipedagConfig):
                # Config objects are not an allowed return type,
                # because they might mess up caching.
                raise TypeError("You can't return a PipedagConfig object from a materializing task.")
            if isinstance(x, ConfigContext):
                # Config objects are not an allowed return type,
                # because they might mess up caching.
                raise TypeError("You can't return a ConfigContext object from a materializing task.")

            # Add missing metadata (if table was not already imperatively materialized)
            if (
                isinstance(x, (Table, RawSql, Blob))
                and id(x) not in task_cache_info.imperative_materialization_state.table_ids
            ):
                if not task.lazy:
                    # task cache_key is output cache_key for eager tables
                    x.cache_key = task_cache_info.cache_key

                x.stage = task.stage

                # Update name:
                # - If no name has been provided, generate on automatically
                # - If the provided name ends with %%, perform name mangling
                object_number = next(auto_suffix_counter)
                auto_suffix = f"{task_cache_info.cache_key}_{object_number:04d}"

                x.name = mangle_table_name(x.name, task.name, auto_suffix)

                if isinstance(x, Table):
                    if x.obj is None:
                        raise TypeError("Underlying table object can't be None")
                    tables.append(x)
                elif isinstance(x, RawSql):
                    if x.sql is None:
                        raise TypeError("Underlying raw sql string can't be None")
                    raw_sqls.append(x)
                elif isinstance(x, Blob):
                    if task.lazy:
                        raise NotImplementedError(
                            "Can't use Blobs with lazy tasks. Invalidation of the"
                            " downstream dependencies is not implemented."
                        )
                    blobs.append(x)
                else:
                    raise NotImplementedError

            return x

        value = deep_map(value, preparation_mutator)
        attach_annotation(task.fn_annotations.get("return"), value)

        return value, tables, raw_sqls, blobs

    @staticmethod
    def _check_names(task: MaterializingTask, tables: list[Table], blobs: list[Blob]):
        if not tables and not blobs:
            # Nothing to check
            return

        # Check names: No duplicates in task
        seen_tn = set()
        seen_bn = set()
        tn_dup = [e.name for e in tables if e.name in seen_tn or seen_tn.add(e.name)]
        bn_dup = [e.name for e in blobs if e.name in seen_bn or seen_bn.add(e.name)]

        if tn_dup or bn_dup:
            raise DuplicateNameError(
                f"Task '{task.name}' returned multiple tables and/or blobs"
                " with the same name.\n"
                f"Duplicate table names: {', '.join(tn_dup) if tn_dup else 'None'}\n"
                f"Duplicate blob names: {', '.join(bn_dup) if bn_dup else 'None'}\n"
                "To enable automatic name mangling,"
                " you can add '%%' at the end of the name."
            )

        # Check names: No duplicates in stage
        ctx = RunContext.get()
        success, tn_dup, bn_dup = ctx.add_names(task.stage, tables, blobs)

        if not success:
            raise DuplicateNameError(
                f"Task '{task.name}' returned tables and/or blobs"
                f" whose name are not unique in the schema '{task.stage}'.\n"
                f"Duplicate table names: {', '.join(tn_dup) if tn_dup else 'None'}\n"
                f"Duplicate blob names: {', '.join(bn_dup) if bn_dup else 'None'}\n"
                "To enable automatic name mangling,"
                " you can add '%%' at the end of the name."
            )

    def _store_task_transaction(
        self,
        task: MaterializingTask,
        tables: list[Table],
        raw_sqls: list[RawSql],
        blobs: list[Blob],
        store_table: Callable[[Table], None],
        store_raw_sql: Callable[[RawSql], None],
        store_blob: Callable[[Blob], None],
        store_metadata: Callable[[], None],
    ) -> list[tuple[Table, Exception]]:
        stage = task.stage
        ctx = RunContext.get()

        stored_tables = []
        stored_blobs = []
        check_failures = []

        try:
            for table in tables:
                ctx.validate_stage_lock(stage)
                try:
                    store_table(table)
                except HookCheckException as e:
                    check_failures.append((table, e))

                stored_tables.append(table)
            for blob in blobs:
                ctx.validate_stage_lock(stage)
                store_blob(blob)
                stored_blobs.append(blob)
            for raw_sql in raw_sqls:
                ctx.validate_stage_lock(stage)
                store_raw_sql(raw_sql)

            ctx.validate_stage_lock(task.stage)
            store_metadata()

        except Exception as e:
            if ConfigContext.get().fail_fast:
                raise e

            # Failed - Roll back everything
            for table in stored_tables:
                self.table_store.delete_table_from_transaction(table)
            for blob in stored_blobs:
                self.blob_store.delete_blob_from_transaction(blob)

            ctx.remove_names(stage, tables, blobs)
            raise e

        return check_failures

    # ### Cache ### #

    def retrieve_cached_output(
        self,
        task: MaterializingTask,
        input_hash: str,
        cache_fn_hash: str,
    ) -> (Materializable, TaskMetadata):
        """Try to retrieve the cached outputs for a task

        :param task: The materializing task for which to retrieve
            the cached output. Must have the `cache_key` attribute set.
        :raises CacheError: if no matching task exists in the cache
        """

        if task.stage.did_commit:
            raise StageError(f"Stage ({task.stage}) already committed.")

        metadata = self.table_store.retrieve_task_metadata(task, input_hash, cache_fn_hash)
        return self.json_decode(metadata.output_json), metadata

    def copy_cached_output_to_transaction_stage(
        self,
        output: Materializable,
        original_metadata: TaskMetadata,
        task: MaterializingTask,
    ):
        """Copy the (non-lazy) outputs from a cached task into the transaction stage

        If the outputs of a task were successfully retrieved from the cache
        using `retrieve_cached_output`, they and the associated metadata
        must be copied from the base stage to the transaction.

        :raises CacheError: if some values in the output can't be found
            in the cache.
        """

        # Get Tables and Blobs from output
        tables = []
        blobs = []
        raw_sqls = []

        def visitor(x):
            if isinstance(x, Table):
                # Tables in external schemas should not get copied
                if x.external_schema is None:
                    tables.append(x)
            elif isinstance(x, RawSql):
                raw_sqls.append(x)
            elif isinstance(x, Blob):
                blobs.append(x)
            return x

        deep_map(output, visitor)

        def store_raw_sql(raw_sql):
            raise Exception("raw sql scripts cannot be part of a non-lazy task")

        # Materialize
        self._check_names(task, tables, blobs)
        self._store_task_transaction(
            task,
            tables,
            raw_sqls,
            blobs,
            self.table_store.copy_table_to_transaction,
            store_raw_sql,
            self.blob_store.copy_blob_to_transaction,
            lambda: self.table_store.store_task_metadata(original_metadata, task.stage),
        )

    def retrieve_most_recent_task_output_from_cache(
        self, task: MaterializingTask, ignore_position_hashes: bool = False
    ) -> (Materializable, TaskMetadata):
        """
        Retrieves the cached output from the most recent execution of a task.

        This retrieval is done based on the name, stage and position hash
        of the task.

        :param task: The materializing task for which to retrieve
            the cached output.
        :param ignore_position_hashes:
            If ``True``, the position hashes of tasks are not checked
            when retrieving the inputs of a task from the cache.
            This simplifies execution of subgraphs if you don't care whether inputs to
            that subgraph are cache invalid. This allows multiple modifications in the
            Graph before the next run updating the cache.
            Attention: This may break automatic cache invalidation.
            And for this to work, any task producing an input
            for the chosen subgraph may never be used more
            than once per stage.
        :return: The output from the task as well as the corresponding metadata.
        :raises CacheError: if no matching task exists in the cache
        """

        # This returns all metadata objects with the same name, stage, AND position
        # hash as `task`. We utilize the position hash to identify a specific
        # task instance, if the same task appears multiple times in a stage.
        metadata = self.table_store.retrieve_all_task_metadata(task, ignore_position_hashes=ignore_position_hashes)

        if not metadata:
            raise CacheError(f"Couldn't find cached output for task '{task}' with matching position hash.")

        # Choose newest entry
        newest_metadata = sorted(metadata, key=lambda m: m.timestamp)[-1]
        cached_output = self.json_decode(newest_metadata.output_json)

        return cached_output, newest_metadata

    # ### Utils ### #

    def json_encode(self, value: Materializable) -> str:
        """Encode a materializable value as json

        In addition to the default types that python can serialise to json,
        this function can also serialise `Table`, `RawSql`, and `Blob` objects.
        The only caveat is, that python's json module doesn't differentiate
        between lists and tuples, which means it is impossible to
        differentiate the two.
        """
        return self.json_encoder.encode(value)

    def json_decode(self, value: str) -> Materializable:
        """Decode a materializable value as json

        Counterpart for the `json_encode` function. Can decode `Table` and
        `Blob` objects.
        """
        return self.json_decoder.decode(value)


def dematerialize_output_from_store(
    store: PipeDAGStore,
    task: Task | TaskGetItem,
    task_output: Materializable,
    as_type: type | None,
) -> Any:
    """
    Dematerializes a task's output from the store

    Must be called inside a ConfigContext and RunContext. Instead
    of a RunContext, a DematerializeRunContext can also be used.
    """
    if isinstance(task, Task):
        root_task = task
    else:
        root_task = task.task

    if as_type is None:
        assert isinstance(root_task, MaterializingTask)
        as_type = root_task.input_type

    run_context = RunContext.get()
    task_output = task.resolve_value(task_output)

    with TaskContext(task):
        return deep_map(
            task_output,
            lambda item: store.dematerialize_item(item, as_type=as_type, ctx=run_context),
        )


def mangle_table_name(table_name: str, task_name: str | None, suffix: str):
    if table_name is None:
        if task_name is None:
            return suffix
        table_name = task_name + "_" + suffix
    elif table_name.endswith("%%"):
        table_name = table_name[:-2] + suffix
    return table_name



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
from pydiverse.common.util.deep_map import deep_map
from pydiverse.common.util.deep_merge import deep_merge
from pydiverse.common.util.disposable import Disposable
from pydiverse.common.util.import_ import requires

from .naming import normalize_name, safe_name

__all__ = [
    "deep_map",
    "deep_merge",
    "Disposable",
    "requires",
    "normalize_name",
    "safe_name",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import ipaddress
import os
import struct
import threading
import uuid
from functools import cached_property
from typing import Any

import msgpack
import pynng
import structlog
from cryptography.fernet import Fernet

from pydiverse.pipedag.errors import IPCError


class IPCClient:
    def __init__(self, addr: str, fernet: Fernet, msg_default=None, msg_ext_hook=None):
        self.logger = structlog.get_logger(logger_name=type(self).__name__)
        self.addr = addr
        self._fernet = fernet
        self.msg_default = msg_default
        self.msg_ext_hook = msg_ext_hook

        self.socket = self._connect()

    def _connect(self):
        self.logger.debug("Opening client connection", addr=self.addr)
        return pynng.Req0(dial=self.addr, resend_time=30_000)

    def request(self, payload: Any) -> Any:
        with self.socket.new_context() as socket:
            self.logger.debug("Client request", payload=payload)
            nonce = uuid.uuid4().bytes[:16]
            msg = msgpack.packb((nonce, payload), default=self.msg_default)
            msg = self._fernet.encrypt(msg)
            socket.send(msg)

            response = socket.recv()
            response = self._fernet.decrypt(response)
            response = msgpack.unpackb(response, use_list=False, ext_hook=self.msg_ext_hook)
            self.logger.debug("Client got response", response=response)
            return response

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["socket"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = self.logger.bind(
            thread=threading.get_ident(),
        )
        self.socket = self._connect()


class IPCServer(threading.Thread):
    """Server for inter process communication

    FORMAT:
        REQUEST: [nonce, payload]
        RESPONSE: [payload]
    """

    def __init__(
        self,
        listen="tcp://127.0.0.1:0",
        msg_default=None,
        msg_ext_hook=None,
    ):
        super().__init__(name="IPCServer", daemon=True)

        # Reduce recv_timeout when running pytest for better performance
        is_running_pytest = "PYDIVERSE_PIPEDAG_PYTEST" in os.environ
        recv_timeout = 200 if not is_running_pytest else 10

        self.socket = pynng.Rep0(listen=listen, recv_timeout=recv_timeout)
        self.nonces = set()

        self._fernet = Fernet(Fernet.generate_key())

        self.__stop_flag = False
        self.logger = structlog.get_logger(logger_name=type(self).__name__, address=self.address)

        self.msg_default = msg_default
        self.msg_ext_hook = msg_ext_hook

    def run(self):
        self.logger.info("Starting IPCServer")
        self.run_loop()

    def stop(self):
        self.__stop_flag = True
        self.logger.debug("Request IPCServer to stop")

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        self.join()

    def run_loop(self):
        context = None
        context_id = 0

        threads = set()
        active_threads = 0

        try:
            context = self.socket.new_context()
            context_id += 1

            while not self.__stop_flag:
                try:
                    data = context.recv()
                    data = self._fernet.decrypt(data)
                    unpacker = msgpack.Unpacker(use_list=False, ext_hook=self.msg_ext_hook)
                    unpacker.feed(data)

                    # Expected: (NONCE, PAYLOAD)
                    if unpacker.read_array_header() != 2:
                        raise IPCError("Expected request to contain exactly two fields: nonce and payload.")

                    nonce = unpacker.unpack()
                    nonce_hex = nonce.hex()

                    if nonce in self.nonces:
                        # Already processing this request
                        self.logger.warning(
                            "Already processing request",
                            nonce=nonce_hex,
                        )

                        context = self.socket.new_context()
                        context_id += 1
                        continue

                    thread_logger = structlog.get_logger(logger_name="IPC Worker", nonce=nonce_hex, id=context_id)
                    thread = threading.Thread(
                        name="IPC Worker",
                        target=self._serve,
                        args=[thread_logger, context, unpacker],
                        daemon=True,
                    )

                    thread.start()
                    threads.add(thread)

                    context = self.socket.new_context()
                    context_id += 1
                except pynng.Timeout:
                    pass
                except Exception:
                    self.logger.exception("Exception occurred in run_loop")

                # Clear list of active threads
                if len(threads) >= active_threads + 25:
                    active_threads = 0
                    for thread in list(threads):
                        if thread.is_alive():
                            active_threads += 1
                        else:
                            self.logger.debug("Joining thread", thread=thread.ident)
                            thread.join()
                            threads.remove(thread)

        finally:
            if context is not None:
                context.close()  # close the open request-response context

            for thread in threads:
                self.logger.debug("Joining thread", thread=thread.ident)
                thread.join()

            self.socket.close()
            self.logger.info("Stopped IPCServer")

    def _serve(self, thread_logger, socket: pynng.Socket, unpacker):
        try:
            msg = unpacker.unpack()
            thread_logger.debug(
                "IPCServer Received",
                message=msg,
                thread=threading.get_ident(),
            )
            reply = self.handle_request(msg)
            thread_logger.debug(
                "IPCServer Reply",
                reply=reply,
                thread=threading.get_ident(),
            )
            reply = msgpack.packb(reply, default=self.msg_default)
            reply = self._fernet.encrypt(reply)
        except Exception as e:
            thread_logger.critical("Uncaught exception in _serve", exc_info=e)
            reply = b""

        try:
            socket.send(reply)
            socket.close()
        except Exception:
            thread_logger.exception("Failed to send reply")

    def handle_request(self, request: dict[str, Any]):
        return None

    @cached_property
    def address(self) -> str:
        address = self.socket.listeners[0].local_address
        addr = address.addr
        port = address.port

        # Convert from big endian to native
        if isinstance(addr, int):
            addr = struct.unpack("=L", addr.to_bytes(4, byteorder="big"))[0]
        port = struct.unpack("=H", port.to_bytes(2, byteorder="big"))[0]

        # Create address string
        ip = ipaddress.ip_address(addr)
        if ip.version == 4:
            return f"tcp://{ip.exploded}:{port}"
        else:
            return f"tcp://[{ip}]:{port}"

    def get_client(self) -> IPCClient:
        return IPCClient(
            addr=self.address,
            fernet=self._fernet,
            msg_default=self.msg_default,
            msg_ext_hook=self.msg_ext_hook,
        )


if __name__ == "__main__":
    with IPCServer() as serv:
        client = serv.get_client()
        client.request(["This is a test"])



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

import pandas as pd
import pytest
import sqlalchemy as sa
from sqlalchemy.exc import ProgrammingError

import tests.util.tasks_library as m
from pydiverse.pipedag import ConfigContext, Flow, Stage, materialize
from pydiverse.pipedag.backend.table.sql.ddl import (
    CreateSchema,
    CreateTableAsSelect,
    CreateViewAsSelect,
    DropTable,
    DropView,
    InsertIntoSelect,
)
from pydiverse.pipedag.container import ExternalTableReference, Schema, Table

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import DATABASE_INSTANCES, skip_instances, with_instances
from tests.util import swallowing_raises
from tests.util.sql import sql_table_expr

pytestmark = [with_instances(DATABASE_INSTANCES)]


@skip_instances("parquet_backend", "parquet_s3_backend")
def test_smoke_table_reference():
    @materialize(version="1.1")
    def in_table():
        table_store = ConfigContext.get().store.table_store
        schema = Schema("user_controlled_schema", prefix="", suffix="")
        table_name = "external_table"
        table_store.execute(CreateSchema(schema, if_not_exists=True))
        table_store.execute(DropTable(table_name, schema, if_exists=True, cascade=True))
        query = sql_table_expr({"col": [0, 1, 2, 3]})
        cmds = (
            [CreateTableAsSelect, InsertIntoSelect]
            if table_store.engine.dialect.name == "ibm_db_sa"
            else [CreateTableAsSelect]
        )
        for cmd in cmds:
            table_store.execute(
                cmd(
                    table_name,
                    schema,
                    query,
                )
            )
        return Table(ExternalTableReference(table_name, schema=schema.get()))

    @materialize(input_type=sa.Table)
    def duplicate_table_reference(tbl: sa.sql.expression.Alias):
        return Table(ExternalTableReference(tbl.original.name, schema=tbl.original.schema))

    with Flow() as f:
        with Stage("sql_table_reference"):
            table1 = in_table()
            table2 = duplicate_table_reference(table1)
            _ = m.assert_table_equal(table1, table2, check_dtype=False)

        with Stage("reference_constant"):
            # This tests ExternalTableReference produced at declare time. It is not
            # recommended to reference any table generated by pipedag code this way.
            # This just helps produce test input for this feature test.
            table3 = Table(ExternalTableReference("external_table", schema="user_controlled_schema"))
            table4 = duplicate_table_reference(table3)
            _ = m.assert_table_equal(table3, table4, check_dtype=False)

    lock_path = Path(__file__).parent / "lock"
    import filelock

    with filelock.FileLock(lock_path):
        assert f.run().successful


try:
    import polars as pl
except ImportError:
    pl = None


@skip_instances("parquet_backend", "parquet_s3_backend")
@pytest.mark.skipif(pl is None, reason="polars is needed for this test")
def test_table_store():
    @materialize(version="1.1")
    def in_table():
        table_store = ConfigContext.get().store.table_store
        schema = Schema("user_controlled_schema", prefix="", suffix="")
        table_name = "external_table"
        table_store.execute(CreateSchema(schema, if_not_exists=True))
        try:
            table_store.execute(DropView("external_view", schema))
        except ProgrammingError:
            pass
        table_store.execute(DropTable(table_name, schema, if_exists=True))
        query = sql_table_expr({"col": [0, 1, 2, 3]})
        cmds = (
            [CreateTableAsSelect, InsertIntoSelect]
            if table_store.engine.dialect.name == "ibm_db_sa"
            else [CreateTableAsSelect]
        )
        for cmd in cmds:
            table_store.execute(
                cmd(
                    table_name,
                    schema,
                    query,
                )
            )
        return Table(ExternalTableReference(table_name, schema=schema.get()))

    @materialize(input_type=sa.Table)
    def duplicate_table_reference(tbl: sa.sql.expression.Alias):
        return Table(ExternalTableReference(tbl.original.name, schema=tbl.original.schema))

    @materialize(version="1.1", input_type=sa.Table)
    def in_view(tbl: sa.sql.expression.Alias):
        table_store = ConfigContext.get().store.table_store
        schema = Schema("user_controlled_schema", prefix="", suffix="")
        view_name = "external_view"
        try:
            # We cannot use if_exists=True here because DB2 does not support it
            table_store.execute(DropView(view_name, schema))
        except ProgrammingError:
            pass
        query = sa.select(tbl.c.col).where(tbl.c.col > 1)
        table_store.execute(
            CreateViewAsSelect(
                view_name,
                schema,
                query,
            )
        )
        return Table(ExternalTableReference(view_name, schema=schema.get()))

    @materialize
    def expected_out_table():
        return Table(
            pd.DataFrame(
                {
                    "col": [0, 1, 2, 3],
                }
            )
        )

    @materialize
    def expected_out_view():
        return Table(
            pd.DataFrame(
                {
                    "col": [2, 3],
                }
            )
        )

    @materialize(input_type=sa.Table)
    def copy_table(tbl: sa.sql.expression.Alias):
        query = sa.select(tbl)
        return Table(query, name=tbl.original.name)

    with Flow() as f:
        with Stage("sql_table_reference"):
            external_table = in_table()
            _ = duplicate_table_reference(external_table)
            expected_external_table = expected_out_table()
            _ = copy_table(external_table)

            _ = m.assert_table_equal(external_table, expected_external_table, check_dtype=False)

        with Stage("sql_view_reference"):
            external_view = in_view(external_table)
            expected_external_view = expected_out_view()
            _ = m.assert_table_equal(external_view, expected_external_view, check_dtype=False)
            external_view_polars = m.noop_polars(external_view)
            external_view_lazy_polars = m.noop_lazy_polars(external_view)
            _ = m.assert_table_equal(external_view_polars, expected_external_view, check_dtype=False)
            _ = m.assert_table_equal(external_view_lazy_polars, expected_external_view, check_dtype=False)

    assert f.run().successful
    assert f.run().successful


def test_bad_table_reference():
    @materialize()
    def bad_table_reference():
        return Table(
            ExternalTableReference(name="this_table_does_not_exist", schema="ext_schema"),
        )

    with Flow() as f:
        with Stage("sql_table_reference"):
            bad_table_reference()

    with swallowing_raises(ValueError, match="this_table_does_not_exist"):
        f.run()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import logging
import tempfile

import sqlalchemy as sa
from sqlalchemy.exc import ProgrammingError

from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import Flow, Stage, Table, materialize
from pydiverse.pipedag.core.config import create_basic_pipedag_config


@materialize(lazy=True)
def lazy_task_1():
    try:
        tbl = Table(sa.text("SELECT-TYPO 1"), name="tbl").materialize()
    except ProgrammingError:
        # This error is expected
        logger = logging.getLogger(__name__ + "-lazy_task_1")
        logger.info("Caught expected error", exc_info=True)

    # now we succeed, but are still not done, yet
    tbl = Table(sa.text("SELECT 'not-done-yet' as a"), name="tbl").materialize()

    # this will create another two tables but they are not returned and won't switch to
    # debug mode
    Table(sa.text("SELECT 3 as a")).materialize()
    Table(sa.text("SELECT 4 as a"), name="tbl2").materialize()

    # now, we succeed with fixing `tbl` and automatically switch in debug mode
    tbl = Table(sa.text("SELECT 1 as a"), name="tbl").materialize()

    # we can also keep a table object:
    tbl_obj = Table(sa.text("SELECT 'not-done-yet' as a"))
    tbl_obj.materialize()

    # this will also automatically switch to debug mode
    tbl_obj.obj = sa.text("SELECT 1 as a")
    tbl_obj.materialize()

    # However, now the flow will stop because cache invalidation cannot deal with debug
    # mode
    return tbl


def main():
    with tempfile.TemporaryDirectory() as temp_dir:
        cfg = create_basic_pipedag_config(
            f"duckdb:///{temp_dir}/db.duckdb",
            disable_stage_locking=True,  # This is special for duckdb
            # Attention: If uncommented, stage and task names might be sent to the
            #   following URL. You can self-host kroki if you like:
            #   https://docs.kroki.io/kroki/setup/install/
            #   You need to install optional dependency 'pydot' for any visualization
            #   URL to appear.
            # kroki_url="https://kroki.io",
        ).get("default")
        with cfg:
            with Flow() as f:
                with Stage("stage_1"):
                    lazy_task_1()

            # Run flow
            f.run()


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish
    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

import pandas as pd

from pydiverse.pipedag import Flow, Stage, materialize
from pydiverse.pipedag.context import ConfigContext
from tests.fixtures.instances import with_instances
from tests.test_raw_sql.util import sql_script


@materialize(input_type=pd.DataFrame)
def raw_sql_object(raw_sql):
    df_1 = raw_sql["table_1"]
    df_2 = raw_sql["table_2"]

    assert not df_1.empty
    assert not df_2.empty


@materialize(input_type=pd.DataFrame)
def raw_sql_individual_table(df_1):
    assert not df_1.empty


@with_instances("postgres", "mssql")
def test_raw_sql_task_input():
    instance_name = ConfigContext.get().instance_name
    dir_ = Path(__file__).parent / "scripts" / instance_name / "create_tables"

    with Flow() as f:
        with Stage("raw_0"):
            simple_tables = sql_script("simple_tables.sql", dir_)

            raw_sql_object(simple_tables)
            raw_sql_individual_table(simple_tables["table_1"])
            raw_sql_individual_table(simple_tables["table_2"])

    f.run()
    f.run()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import sqlalchemy as sa

from pydiverse.pipedag import ConfigContext, Flow, Stage, StageLockContext, materialize
from pydiverse.pipedag.backend.table.sql.ddl import (
    CreateSchema,
    CreateTableAsSelect,
    DropTable,
    InsertIntoSelect,
)
from pydiverse.pipedag.container import ExternalTableReference, RawSql, Schema, Table
from tests.fixtures.instances import with_instances

# Parameterize all tests in this file with several instance_id configurations
from tests.util.sql import sql_table_expr


@with_instances("postgres")
def test_external_table_inputs():
    @materialize(version="1.1")
    def make_external_table():
        table_store = ConfigContext.get().store.table_store
        schema = Schema("external_schema", prefix="", suffix="")
        table_name = "external_table"
        table_store.execute(CreateSchema(schema, if_not_exists=True))
        table_store.execute(DropTable(table_name, schema, if_exists=True))
        query = sql_table_expr({"col": [0, 1, 2, 3]})
        cmds = (
            [CreateTableAsSelect, InsertIntoSelect]
            if table_store.engine.dialect.name == "ibm_db_sa"
            else [CreateTableAsSelect]
        )
        for cmd in cmds:
            table_store.execute(
                cmd(
                    table_name,
                    schema,
                    query,
                )
            )
        return Table(ExternalTableReference(table_name, schema=schema.get()))

    @materialize(input_type=sa.Table)
    def duplicate_table_reference():
        return Table(pd.DataFrame())

    @materialize(input_type=pd.DataFrame)
    def identity(table: pd.DataFrame):
        return table

    with Flow() as f:
        with Stage("sql_table_origin"):
            _ = make_external_table()

        with Stage("sql_table_linked"):
            table = duplicate_table_reference()
            output = identity(table)

    with StageLockContext():
        # Normal execution. duplicate_table_reference should return an empty table
        result = f.run()
        assert result.get(output, as_type=pd.DataFrame).shape[0] == 0

    with StageLockContext():
        # Linked execution.
        # Input of identity should be the output of make_external_table.
        result = f.run(
            inputs={
                table: ExternalTableReference(
                    "external_table",
                    schema="external_schema",
                )
            }
        )
        assert result.get(output, as_type=pd.DataFrame).shape[0] == 4


@with_instances("mssql")
def test_external_table_inputs_rawsql():
    @materialize(version="1.1")
    def make_external_table():
        table_store = ConfigContext.get().store.table_store
        schema = Schema("external_schema", prefix="", suffix="")
        table_name = "external_table"
        table_store.execute(CreateSchema(schema, if_not_exists=True))
        table_store.execute(DropTable(table_name, schema, if_exists=True))
        query = sql_table_expr({"col": [0, 1, 2, 3]})
        cmds = (
            [CreateTableAsSelect, InsertIntoSelect]
            if table_store.engine.dialect.name == "ibm_db_sa"
            else [CreateTableAsSelect]
        )
        for cmd in cmds:
            table_store.execute(
                cmd(
                    table_name,
                    schema,
                    query,
                )
            )
        return Table(ExternalTableReference(table_name, schema=schema.get()))

    @materialize(input_type=sa.Table)
    def duplicate_table_reference():
        sql = """
        SELECT 1 as col INTO sql_table_linked__tmp.duplicate_table_reference
        """
        return RawSql(sql)

    @materialize(input_type=pd.DataFrame)
    def identity(table: pd.DataFrame):
        return table

    with Flow() as f:
        with Stage("sql_table_origin"):
            _ = make_external_table()

        with Stage("sql_table_linked"):
            table = duplicate_table_reference()["duplicate_table_reference"]
            output = identity(table)

    with StageLockContext():
        # Normal execution. duplicate_table_reference should return an empty table
        result = f.run()
        assert result.get(output, as_type=pd.DataFrame).shape[0] == 1

    with StageLockContext():
        # Linked execution.
        # Input of identity should be the output of make_external_table.
        result = f.run(
            inputs={
                table.task: ExternalTableReference(
                    "external_table",
                    schema="external_schema",
                )
            }
        )
        assert result.get(output, as_type=pd.DataFrame).shape[0] == 4


@with_instances("postgres")
def test_external_table_inputs_nout():
    @materialize(version="1.1")
    def make_external_table():
        table_store = ConfigContext.get().store.table_store
        schema = Schema("external_schema", prefix="", suffix="")
        table_name = "external_table"
        table_store.execute(CreateSchema(schema, if_not_exists=True))
        table_store.execute(DropTable(table_name, schema, if_exists=True))
        query = sql_table_expr({"col": [0, 1, 2, 3]})
        cmds = (
            [CreateTableAsSelect, InsertIntoSelect]
            if table_store.engine.dialect.name == "ibm_db_sa"
            else [CreateTableAsSelect]
        )
        for cmd in cmds:
            table_store.execute(
                cmd(
                    table_name,
                    schema,
                    query,
                )
            )
        return Table(ExternalTableReference(table_name, schema=schema.get()))

    @materialize(input_type=sa.Table, nout=2)
    def duplicate_table_reference():
        return Table(pd.DataFrame([1])), Table(pd.DataFrame())

    @materialize(input_type=pd.DataFrame)
    def identity(table: pd.DataFrame):
        return table

    with Flow() as f:
        with Stage("sql_table_origin"):
            _ = make_external_table()

        with Stage("sql_table_linked"):
            table, dummy = duplicate_table_reference()
            output = identity(table)

    with StageLockContext():
        # Normal execution. duplicate_table_reference should return an empty table
        result = f.run()
        assert result.get(output, as_type=pd.DataFrame).shape[0] == 1

    with StageLockContext():
        # Linked execution.
        # Input of identity should be the output of make_external_table.
        result = f.run(
            inputs={
                table.task: ExternalTableReference(
                    "external_table",
                    schema="external_schema",
                )
            }
        )
        assert result.get(output, as_type=pd.DataFrame).shape[0] == 4


@with_instances("postgres")
def test_external_table_inputs_no_run():
    @materialize(version="1.1")
    def make_external_table():
        table_store = ConfigContext.get().store.table_store
        schema = Schema("external_schema", prefix="", suffix="")
        table_name = "external_table"
        table_store.execute(CreateSchema(schema, if_not_exists=True))
        table_store.execute(DropTable(table_name, schema, if_exists=True))
        query = sql_table_expr({"col": [0, 1, 2, 3]})
        cmds = (
            [CreateTableAsSelect, InsertIntoSelect]
            if table_store.engine.dialect.name == "ibm_db_sa"
            else [CreateTableAsSelect]
        )
        for cmd in cmds:
            table_store.execute(
                cmd(
                    table_name,
                    schema,
                    query,
                )
            )
        return Table(ExternalTableReference(table_name, schema=schema.get()))

    @materialize(input_type=sa.Table)
    def duplicate_table_reference():
        return Table(pd.DataFrame())

    @materialize(input_type=pd.DataFrame)
    def identity(table: pd.DataFrame):
        return table

    with Flow() as f:
        with Stage("sql_table_origin"):
            _ = make_external_table()

        with Stage("sql_table_linked"):
            table = duplicate_table_reference()
            output = identity(table)

    with StageLockContext():
        # Normal execution. duplicate_table_reference should return an empty table
        result = f.run()
        assert result.get(output, as_type=pd.DataFrame).shape[0] == 0

    with StageLockContext():
        # Linked execution.
        # Input of identity should be the output of make_external_table.
        result = f.run(
            run_selection=[output],
            inputs={
                table: ExternalTableReference(
                    "external_table",
                    schema="external_schema",
                )
            },
        )
        assert result.get(output, as_type=pd.DataFrame).shape[0] == 4



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import itertools
import os
import signal
import time
from multiprocessing import Process, Queue
from queue import Empty
from threading import Thread

import pytest
from _pytest.python import Config

from .util import parse_config
from .worker import start_worker


class Session:
    def __init__(self, config: Config):
        self.config = config
        self.msg_queue = Queue()
        self.work_queue = Queue()

        self.workers = []
        self.worker_id_counter = itertools.count()
        self.running_workers = set()

        self.debug_worker_group = {}
        self.debug_worker_test = {}

        self._should_shutdown = False
        self._shutdown_reason = "None"

        # Hide path from output
        reporter = config.pluginmanager.getplugin("terminalreporter")
        reporter.showfspath = False

        # Get the number of workers
        try:
            num_workers = parse_config(config, "workers")
            if num_workers == "auto":
                num_workers = os.cpu_count() or 1
            elif num_workers is not None:
                num_workers = int(num_workers)
            else:
                num_workers = 1
            self.num_workers = num_workers
        except ValueError:
            raise ValueError('workers can only be an integer or "auto"') from None

        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def pytest_runtestloop(self, session):
        if session.testsfailed and not session.config.option.continue_on_collection_errors:
            raise session.Interrupted(
                f"{session.testsfailed} error{'s' if session.testsfailed != 1 else ''} during collection"
            )

        if session.config.option.collectonly:
            return True

        # Prepare everything
        grouped_items = self.get_grouped_items(session)
        num_workers = min(len(grouped_items), self.num_workers)

        self.prepare_work_queue(num_workers, grouped_items)
        self.start_workers(num_workers)

        # Monitor workers
        def check_workers_alive_loop():
            while True:
                self.check_workers_alive()
                time.sleep(1)

        thread = Thread(target=check_workers_alive_loop, daemon=True)
        thread.start()

        # Handle messages
        while self.running_workers and not self._should_shutdown:
            try:
                msg, kwargs = self.msg_queue.get(timeout=1)
            except Empty:
                continue

            worker = self.workers[kwargs["worker_id"]]

            if msg == "sessionstart":
                pass
            elif msg == "sessionfinish":
                worker.join()
                self.running_workers.remove(worker)
            elif msg == "logstart":
                self.config.hook.pytest_runtest_logstart(nodeid=kwargs["nodeid"], location=kwargs["location"])
            elif msg == "logfinish":
                self.config.hook.pytest_runtest_logfinish(nodeid=kwargs["nodeid"], location=kwargs["location"])
            elif msg == "logreport":
                report = self.config.hook.pytest_report_from_serializable(config=self.config, data=kwargs["report"])
                self.config.hook.pytest_runtest_logreport(report=report)
            elif msg == "DEBUG_start_group":
                self.debug_worker_group[worker] = kwargs["group_name"]
            elif msg == "DEBUG_start_test":
                self.debug_worker_test[worker] = kwargs["nodeid"]
            else:
                error_msg = f"Unknown message: {msg}"
                raise ValueError(error_msg)

        if self._should_shutdown:
            pytest.exit(self._shutdown_reason)

        return True

    def get_grouped_items(self, session) -> dict[str, list]:
        groupings = self.config.hook.pytest_parallelize_group_items(
            config=self.config,
            items=session.items,
        )

        if len(groupings) == 0:
            msg = "No implementation for `pytest_parallelize_group_items` hook found."
            raise AssertionError(msg)
        if len(groupings) > 1:
            msg = "Multiple implementations for `pytest_parallelize_group_items` hook found."
            raise AssertionError(msg)

        grouping = groupings[0]

        reporter = session.config.pluginmanager.getplugin("terminalreporter")
        if reporter.showheader and not reporter.no_header:
            reporter.line(
                f"Split tests into {len(grouping)} parallelization groups: " + ", ".join(sorted(list(grouping)))
            )

        return grouping

    def prepare_work_queue(self, num_workers, grouped_items):
        for group, items in grouped_items.items():
            node_ids = [item.nodeid for item in items]
            self.work_queue.put(("GROUP", (group, node_ids)))

        for _ in range(num_workers):
            self.work_queue.put(("STOP", ()))

    def start_workers(self, num_workers):
        args = self.config.args
        option_dict = vars(self.config.option)

        for _ in range(num_workers):
            worker_id = next(self.worker_id_counter)
            worker = Process(
                target=start_worker,
                name=f"pytest-worker-{worker_id:03}",
                args=(
                    worker_id,
                    self.work_queue,
                    self.msg_queue,
                    args,
                    option_dict,
                ),
                daemon=False,
            )
            worker.start()

            self.workers.append(worker)
            self.running_workers.add(worker)

    def check_workers_alive(self):
        for worker in self.running_workers:
            if not worker.is_alive():
                group_name = self.debug_worker_group.get(worker)
                test_name = self.debug_worker_test.get(worker)
                msg = (
                    f"Worker {worker.name} died with exit code {worker.exitcode}."
                    f" (group = {group_name}, test = {test_name})"
                )
                self.shutdown(msg)

    def shutdown(self, reason: str):
        self._shutdown_reason = reason
        self._should_shutdown = True

    def exit_gracefully(self, signum, frame):
        for worker in self.workers:
            worker.terminate()

        signame = signal.Signals(signum).name
        pytest.exit(f"Received signal {signame}")
